# -*- coding: utf-8 -*-

"""Resources for y0."""

import os

HERE = os.path.abspath(os.path.dirname(__file__))
ASIA_PATH = os.path.join(HERE, "asia.csv")
VIRAL_PATHOGENESIS_PATH = os.path.join(HERE, "viral_pathogenesis.json")



# -*- coding: utf-8 -*-

"""Operations for mutating and simplifying expressions."""

from ..dsl import (
    Distribution,
    Fraction,
    OrderingHint,
    P,
    Probability,
    Product,
    ensure_ordering,
)

__all__ = [
    "chain_expand",
    "fraction_expand",
    "bayes_expand",
]


def chain_expand(
    p: Probability, *, reorder: bool = True, ordering: OrderingHint = None
) -> Product:
    r"""Expand a probability distribution to a product of conditional probabilities on single variables.

    :param p: The given probability expression
    :param reorder: Should the variables be reordered with respect to the ordering before expanding? This is important
        because there are a variety of equivalent expressions that can't be directly matched.
    :param ordering: An ordering to be used if ``reorder`` is true. If none, automatically generates a canonical
        ordering using :func:`y0.dsl.ensure_ordering`.
    :return: A product representing the expanded distribution, in which each probability term is a markov kernel

    :raises ValueError: if the ordering is passed explicitly and it does not cover all variables

    Two variables:

    .. math::
        P(X,Y)=P(X|Y)*P(Y)

    >>> from y0.dsl import P, X, Y
    >>> assert chain_expand(P(X | Y)) == P(X | Y) * P(Y)

    The recurrence relation for many variables is defined as:

    .. math::
        P(X_n,\dots,X_1) = P(X_n|X_{n-1},\dots,X_1) \times P(X_{n-1},\dots,X_1)

    >>> from y0.dsl import P, A, X, Y, Z
    >>> chain_expand(P(X, Y, Z)) == P(X | (Y, Z)) * P(Y | Z) * P(Z)

    Extra conditions come along for the ride.

    >>> chain_expand(P(X & Y | A)) == P(X | (Y, Z, A)) * P(Y | (Z, A)) * P(Z | A)
    """
    if reorder:
        _ordering = ensure_ordering(p, ordering=ordering)
        if any(v not in _ordering for v in p.distribution.children):
            raise ValueError
        ordered_children = tuple(v for v in _ordering if v in p.distribution.children)
    else:
        ordered_children = p.distribution.children

    return Product(
        tuple(
            P(
                Distribution(children=(ordered_children[i],)).given(
                    ordered_children[i + 1 :] + p.distribution.parents
                )
            )
            for i in range(len(ordered_children))
        )
    )


def fraction_expand(p: Probability) -> Fraction:
    r"""Expand a probability distribution with fractions.

    :param p: The given probability expression
    :returns: A fraction representing the joint distribution

    The simple case has one child variable ($A$) and one parent variable ($B$):

    .. math::
        P(A | B) = \frac{P(A,B)}{P(B)}

    In general, with many children $Y_i$ and many parents $X_i$:

    .. math::
        P(Y_1,\dots,Y_n | X_1, \dots, X_m) = \frac{P(Y_1,\dots,Y_n,X_1,\dots,X_m)}{P(X_1,\dots,X_m)}
    """
    return p.uncondition() / P(p.distribution.parents)


def bayes_expand(p: Probability) -> Fraction:
    r"""Expand a probability distribution using Bayes' theorem.

    :param p: The given probability expression, with arbitrary number of children and parents
    :returns: A fraction representing the joint distribution

    .. math::
        P(Y_1,\dots,Y_n|X_1,\dots,X_m)
        = \frac{P(Y_1,\dots,Y_n,X_1,\dots,X_m)}{\sum_{Y_1,\dots,Y_n} P(Y_1,\dots,Y_n,X_1,\dots,X_m)}

    .. note:: This expansion will create a different but equal expression to :func:`fraction_expand`.
    """
    return p.uncondition().marginalize(p.distribution.children)



# -*- coding: utf-8 -*-

"""Graph data structures."""

from __future__ import annotations

import itertools as itt
import json
from dataclasses import dataclass, field
from typing import (
    Any,
    Collection,
    Generic,
    Iterable,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import networkx as nx
from ananke.graphs import ADMG
from networkx.utils import open_file

__all__ = [
    "NxMixedGraph",
    "CausalEffectGraph",
    "DEFULT_PREFIX",
    "DEFAULT_TAG",
    "admg_to_latent_variable_dag",
    "admg_from_latent_variable_dag",
    "set_latent",
]

X = TypeVar("X")
CausalEffectGraph = Any

#: The default key in a latent variable DAG represented as a :class:`networkx.DiGraph`
#: for nodes that corresond to "latent" variables
DEFAULT_TAG = "hidden"
#: The default prefix for latent variables in a latent variable DAG represented. After the prefix,
#: there will be a number assigned that's incremented during construction.
DEFULT_PREFIX = "u_"


@dataclass
class NxMixedGraph(Generic[X]):
    """A mixed graph based on a :class:`networkx.Graph` and a :class:`networkx.DiGraph`.

    Example usage:

    .. code-block:: python

        graph = NxMixedGraph()
        graph.add_directed_edge('X', 'Y')
        graph.add_undirected_edge('X', 'Y')

        # Convert to an Ananke acyclic directed mixed graph
        admg_graph = graph.to_admg()
    """

    #: A directed graph
    directed: nx.DiGraph = field(default_factory=nx.DiGraph)
    #: A undirected graph
    undirected: nx.Graph = field(default_factory=nx.Graph)

    def add_directed_edge(self, u: X, v: X, **attr) -> None:
        """Add a directed edge from u to v."""
        self.directed.add_edge(u, v, **attr)
        self.undirected.add_node(u)
        self.undirected.add_node(v)

    def add_undirected_edge(self, u: X, v: X, **attr) -> None:
        """Add an undirected edge between u and v."""
        self.undirected.add_edge(u, v, **attr)
        self.directed.add_node(u)
        self.directed.add_node(v)

    def to_admg(self) -> ADMG:
        """Get an ADMG instance."""
        di_edges = list(self.directed.edges())
        bi_edges = list(self.undirected.edges())
        vertices = list(
            self.directed
        )  # could be either since they're maintained together
        return ADMG(vertices=vertices, di_edges=di_edges, bi_edges=bi_edges)

    @classmethod
    def from_admg(cls, admg: ADMG) -> NxMixedGraph:
        """Create from an ADMG."""
        return cls.from_edges(directed=admg.di_edges, undirected=admg.bi_edges)

    def to_latent_variable_dag(
        self,
        *,
        prefix: Optional[str] = None,
        start: int = 0,
        tag: Optional[str] = None,
    ) -> nx.DiGraph:
        """Create a labeled DAG where bi-directed edges are assigned as nodes upstream of their two incident nodes.

        :param prefix: The prefix for latent variables. If none, defaults to :data:`y0.graph.DEFAULT_PREFIX`.
        :param start: The starting number for latent variables (defaults to 0, could be changed to 1 if desired)
        :param tag: The key for node data describing whether it is latent.
            If None, defaults to :data:`y0.graph.DEFAULT_TAG`.
        :return: A latent variable DAG.
        """
        return _latent_dag(
            di_edges=self.directed.edges(),
            bi_edges=self.undirected.edges(),
            prefix=prefix,
            start=start,
            tag=tag,
        )

    @classmethod
    def from_latent_variable_dag(
        cls, graph: nx.DiGraph, tag: Optional[str] = None
    ) -> NxMixedGraph:
        """Load a labeled DAG."""
        if tag is None:
            tag = DEFAULT_TAG
        if any(tag not in data for data in graph.nodes.values()):
            raise ValueError(f"missing label {tag} in one or more nodes.")

        rv = cls()
        for node, data in graph.nodes.items():
            if data[tag]:
                for a, b in itt.combinations(graph.successors(node), 2):
                    rv.add_undirected_edge(a, b)
            else:
                for child in graph.successors(node):
                    rv.add_directed_edge(node, child)
        return rv

    def to_causaleffect(self) -> CausalEffectGraph:
        """Get a causaleffect R object.

        :returns: A causaleffect R object.

        .. warning:: Appropriate R imports need to be done first for 'causaleffect' and 'igraph'.
        """
        import rpy2.robjects

        return rpy2.robjects.r(self.to_causaleffect_str())

    def draw(self, ax=None, title=None):
        """Render the graph using matplotlib.

        :param ax: Axis to draw on (if none specified, makes a new one)
        :param title: The optional title to show with the graph
        """
        joint = nx.MultiGraph()
        joint.add_edges_from(self.directed.edges)
        joint.add_edges_from(self.undirected.edges)
        layout = nx.nx_pydot.graphviz_layout(joint, prog="dot")

        u_proxy = nx.DiGraph()
        u_proxy.add_edges_from(self.undirected.edges)

        if ax is None:
            import matplotlib.pyplot as plt

            ax = plt.gca()

        nx.draw_networkx_nodes(self.directed, pos=layout, ax=ax)
        nx.draw_networkx_labels(self.directed, pos=layout, ax=ax)
        nx.draw_networkx_edges(self.directed, pos=layout, edge_color="b", ax=ax)
        nx.draw_networkx_edges(
            u_proxy,
            pos=layout,
            ax=ax,
            connectionstyle="arc3, rad=0.2",
            arrowstyle="-",
            edge_color="r",
        )

        if title:
            ax.set_title(title)
        ax.axis("off")

    @classmethod
    def from_causaleffect(cls, graph) -> NxMixedGraph:
        """Construct an instance from a causaleffect R graph."""
        raise NotImplementedError

    def to_causaleffect_str(self) -> str:
        """Get a string to be imported by R."""
        if not self.directed:
            raise ValueError("graph must have some directed edges")

        formula = ", ".join(f"{u} -+ {v}" for u, v in self.directed.edges())
        if self.undirected:
            formula += "".join(
                f", {u} -+ {v}, {v} -+ {u}" for u, v in self.undirected.edges()
            )

        rv = f"g <- graph.formula({formula}, simplify = FALSE)"
        for i in range(self.undirected.number_of_edges()):
            idx = 2 * i + self.directed.number_of_edges() + 1
            rv += (
                f'\ng <- set.edge.attribute(graph = g, name = "description",'
                f' index = c({idx}, {idx + 1}), value = "U")'
            )

        return rv

    @classmethod
    def from_edges(
        cls,
        directed: Iterable[Tuple[X, X]],
        undirected: Optional[Iterable[Tuple[X, X]]] = None,
    ) -> NxMixedGraph:
        """Make a mixed graph from a pair of edge lists."""
        rv = cls()
        for u, v in directed:
            rv.add_directed_edge(u, v)
        for u, v in undirected or []:
            rv.add_undirected_edge(u, v)
        return rv

    @classmethod
    def from_adj(
        cls,
        directed: Mapping[X, Collection[X]],
        undirected: Mapping[X, Collection[X]],
    ) -> NxMixedGraph:
        """Make a mixed graph from a pair of adjacency lists."""
        rv = cls()
        for u, vs in directed.items():
            for v in vs:
                rv.add_directed_edge(u, v)
        for u, vs in undirected.items():
            for v in vs:
                rv.add_undirected_edge(u, v)
        return rv

    @classmethod
    @open_file(1)
    def from_causalfusion_path(cls, file) -> NxMixedGraph:
        """Load a graph from a CausalFusion JSON file."""
        return cls.from_causalfusion_json(json.load(file))

    @classmethod
    def from_causalfusion_json(cls, data: Mapping[str, Any]) -> NxMixedGraph:
        """Load a graph from a CausalFusion JSON object."""
        rv = cls()
        for edge in data["edges"]:
            u, v = edge["from"], edge["to"]
            if edge["type"] == "directed":
                rv.add_directed_edge(u, v)
            elif edge["type"] == "bidirected":
                rv.add_undirected_edge(u, v)
            else:
                raise ValueError(f'unhandled edge type: {edge["type"]}')
        return rv


def admg_to_latent_variable_dag(
    graph: ADMG,
    prefix: Optional[str] = None,
    start: int = 0,
    tag: Optional[str] = None,
) -> nx.DiGraph:
    """Convert an ADMG to a latent variable DAG.

    :param graph: An ADMG
    :param prefix: The prefix for latent variables. If none, defaults to :data:`y0.graph.DEFAULT_PREFIX`.
    :param start: The starting number for latent variables (defaults to 0, could be changed to 1 if desired)
    :param tag: The key for node data describing whether it is latent.
        If None, defaults to :data:`y0.graph.DEFAULT_TAG`.
    :return: A latent variable DAG.
    """
    return _latent_dag(
        graph.di_edges,
        graph.bi_edges,
        prefix=prefix,
        start=start,
        tag=tag,
    )


def admg_from_latent_variable_dag(
    graph: nx.DiGraph, *, tag: Optional[str] = None
) -> ADMG:
    """Convert a latent variable DAG to an ADMG.

    :param graph: A latent variable directed acyclic graph (LV-DAG)
    :param tag: The key for node data describing whether it is latent.
        If None, defaults to :data:`y0.graph.DEFAULT_TAG`.
    :return: An ADMG
    """
    return NxMixedGraph.from_latent_variable_dag(graph, tag=tag).to_admg()


def _latent_dag(
    di_edges: Iterable[Tuple[str, str]],
    bi_edges: Iterable[Tuple[str, str]],
    *,
    prefix: Optional[str] = None,
    start: int = 0,
    tag: Optional[str] = None,
) -> nx.DiGraph:
    """Create a labeled DAG where bi-directed edges are assigned as nodes upstream of their two incident nodes.

    :param di_edges: A list of directional edges
    :param bi_edges: A list of bi-directional edges
    :param prefix: The prefix for latent variables. If none, defaults to :data:`y0.graph.DEFAULT_PREFIX`.
    :param start: The starting number for latent variables (defaults to 0, could be changed to 1 if desired)
    :param tag: The key for node data describing whether it is latent.
        If None, defaults to :data:`y0.graph.DEFAULT_TAG`.
    :return: A latent variable DAG.
    """
    if tag is None:
        tag = DEFAULT_TAG
    if prefix is None:
        prefix = DEFULT_PREFIX

    rv = nx.DiGraph()
    rv.add_nodes_from(itt.chain.from_iterable(bi_edges))
    rv.add_edges_from(di_edges)
    nx.set_node_attributes(rv, False, tag)
    for i, (u, v) in enumerate(sorted(bi_edges), start=start):
        latent_node = f"{prefix}{i}"
        rv.add_node(latent_node, **{tag: True})
        rv.add_edge(latent_node, u)
        rv.add_edge(latent_node, v)
    return rv


def set_latent(
    graph: nx.DiGraph,
    latent_nodes: Union[str, Iterable[str]],
    tag: Optional[str] = None,
) -> None:
    """Quickly set the latent variables in a graph."""
    if tag is None:
        tag = DEFAULT_TAG
    if isinstance(latent_nodes, str):
        latent_nodes = [latent_nodes]

    latent_nodes = set(latent_nodes)
    for node, data in graph.nodes(data=True):
        data[tag] = node in latent_nodes



# -*- coding: utf-8 -*-

"""A parser for y0 internal DSL probability expressions based on Python's :func:`eval` function."""

import string

from y0.dsl import Expression, P, Q, Sum, Variable

__all__ = [
    "parse_y0",
]

LOCALS = {
    "P": P,
    "PROB": P,
    "Prob": P,
    "PROBABILITY": P,
    "Probability": P,
    "SUM": Sum,
    "Sum": Sum,
    "Q": Q,
    "QFactor": Q,
}

for letter in string.ascii_uppercase:
    if letter in {"P", "Q"}:
        continue
    LOCALS[letter] = Variable(letter)
    for index in range(10):
        name = f"{letter}{index}"
        LOCALS[name] = Variable(name)

        name_underscored = f"{letter}_{index}"
        LOCALS[name_underscored] = Variable(name_underscored)


def parse_y0(s) -> Expression:
    """Parse a valid Python expression using the :mod:`y0.dsl` objects, written in a string.

    :param s: The string to parse. Should be a valid Python expression given ``from y0.dsl import *``.
        Variables of the form A-Z, A0-Z0, A_0-Z_0, A1-Z1, A_1-Z_1, ..., A9-Z9, A_9-Z_9 are available.
    :return: An expression object.

    >>> from y0.parser import parse_y0
    >>> from y0.dsl import P, A, B, Sum
    >>> parse_y0('Sum[B](P(A|B) * P(B))') == Sum[B](P(A|B) * P(B))
    True
    """
    return eval(s, {}, LOCALS)  # noqa:S307



# -*- coding: utf-8 -*-

"""An implementation to get conditional independencies of an ADMG."""

import copy
from functools import partial
from itertools import chain, combinations, groupby
from typing import Iterable, List, Optional, Set, Tuple, Union

import networkx as nx
from ananke.graphs import ADMG, SG
from tqdm import tqdm

from ..graph import NxMixedGraph
from ..struct import DSeparationJudgement
from ..util.combinatorics import powerset

__all__ = [
    "are_d_separated",
    "minimal",
    "get_conditional_independencies",
]


def get_conditional_independencies(
    graph: Union[NxMixedGraph, SG],
    *,
    policy=None,
    **kwargs,
) -> Set[DSeparationJudgement]:
    """Get the conditional independencies from the given ADMG.

    Conditional independencies is the minmal set of d-separation judgements to cover
    the unique left/right combinations in all valid d-separation.

    :param graph: An acyclic directed mixed graph
    :param policy: Retention policy when more than one conditional independency option exists (see minimal for details)
    :param kwargs: Other keyword arguments are passed to d_separations
    :return: A set of conditional dependencies

    .. seealso:: Original issue https://github.com/y0-causal-inference/y0/issues/24
    """
    if isinstance(graph, NxMixedGraph):
        graph = graph.to_admg()
    if policy is None:
        policy = topological_policy(graph)
    return minimal(
        d_separations(graph, **kwargs),
        policy=policy,
    )


def minimal(
    judgements: Iterable[DSeparationJudgement], policy=None
) -> Set[DSeparationJudgement]:
    """Given some d-separations, reduces to a 'minimal' collection.

    For indepdencies of the form A _||_ B | {C1, C2, ...} the minimal collection will::

    - Have only one independency with the same A/B nodes.
    - If there are multiples sets of C-nodes, the kept d-separation will be the first/minimal
      element in the group sorted according to `policy` argument.

    The default policy is to sort by the shortest set of conditions & then lexicographic.

    :param judgements: Collection of judgements to minimize
    :param policy: Function from d-separation to a representation suitable for sorting.
    :return: A set of judgements that is minimal (as described above)
    """
    if policy is None:
        policy = _len_lex
    judgements = sorted(judgements, key=_judgement_grouper)
    return {min(vs, key=policy) for k, vs in groupby(judgements, _judgement_grouper)}


def topological_policy(graph: ADMG):
    """Sort d-separations by condition length and topological order.

    This policy will prefers small collections, and collections with variables earlier
    in topological order for collections of the same size.

    :param graph: ADMG
    :return: A function suitable for use as a sort key on d-separations
    """
    order = graph.topological_sort()
    return partial(_topological_policy, order=order)


def _topological_policy(judgement: DSeparationJudgement, order):
    return (
        len(judgement.conditions),
        sum((order.index(v) for v in judgement.conditions)),
    )


def _judgement_grouper(judgement: DSeparationJudgement) -> Tuple[str, str]:
    """Simplify d-separation to just left & right element (for grouping left/right pairs)."""
    return judgement.left, judgement.right


def _len_lex(judgement: DSeparationJudgement) -> Tuple[int, str]:
    """Sort by length of conditions & the lexicography a d-separation."""
    return len(judgement.conditions), ",".join(judgement.conditions)


def disorient(graph: SG) -> nx.Graph:
    """Convert an :mod:`ananke` mixed directed/undirected into a undirected (networkx) graph."""
    rv = nx.Graph()
    rv.add_nodes_from(graph.vertices)
    rv.add_edges_from(chain(graph.di_edges, graph.ud_edges, graph.bi_edges))
    return rv


def get_moral_links(graph: SG) -> List[Tuple[str, str]]:
    """Generate links to ensure all co-parents in a graph are linked.

    May generate links that already exist as we assume we are not working on a multi-graph.

    :param graph: Graph to process
    :return: An collection of edges to add.
    """
    parents = [graph.parents([v]) for v in graph.vertices]
    moral_links = [
        *chain(*[combinations(nodes, 2) for nodes in parents if len(parents) > 1])
    ]
    return moral_links


def are_d_separated(
    graph: SG, a: str, b: str, *, conditions: Optional[Iterable[str]] = None
) -> DSeparationJudgement:
    """Test if nodes named by a & b are d-separated in G.

    a & b can be provided in either order and the order of conditions does not matter.
    However DSeparationJudgement may put things in canonical order.

    :param graph: Graph to test
    :param a: A node in the graph
    :param b: A node in the graph
    :param conditions: A collection of graph nodes
    :return: T/F and the final graph (as evidence)
    """
    conditions = set(conditions) if conditions else set()
    named = {a, b}.union(conditions)

    # Filter to ancestors
    keep = graph.ancestors(named)
    graph = copy.deepcopy(graph.subgraph(keep))

    # Moralize (link parents of mentioned nodes)
    for u, v in get_moral_links(graph):
        graph.add_udedge(u, v)

    # disorient & remove conditions
    evidence_graph = disorient(graph)

    keep = set(evidence_graph.nodes) - set(conditions)
    evidence_graph = evidence_graph.subgraph(keep)

    # check for path....
    separated = not nx.has_path(evidence_graph, a, b)  # If no path, then d-separated!

    return DSeparationJudgement.create(
        left=a, right=b, conditions=conditions, separated=separated
    )


def d_separations(
    graph: Union[NxMixedGraph, SG],
    *,
    max_conditions: Optional[int] = None,
    verbose: Optional[bool] = False,
    return_all: Optional[bool] = False,
) -> Iterable[DSeparationJudgement]:
    """Generate d-separations in the provided graph.

    :param graph: Graph to search for d-separations.
    :param max_conditions: Longest set of conditions to investigate
    :param return_all: If false (default) only returns the first d-separation per left/right pair.
    :param verbose: If true, prints extra output with tqdm
    :yields: True d-separation judgements
    """
    if isinstance(graph, NxMixedGraph):
        graph = graph.to_admg()

    vertices = set(graph.vertices)
    for a, b in tqdm(
        combinations(vertices, 2), disable=not verbose, desc="d-separation check"
    ):
        for conditions in powerset(vertices - {a, b}, stop=max_conditions):
            judgement = are_d_separated(graph, a, b, conditions=conditions)
            if judgement.separated:
                yield judgement
                if not return_all:
                    break



# -*- coding: utf-8 -*-

"""Parsers for various probability expression grammars."""

from .ce.grammar import parse_causaleffect  # noqa:F401
from .craig.grammar import parse_craig  # noqa:F401
from .internal import parse_y0  # noqa:F401



# -*- coding: utf-8 -*-

"""Test the causaleffect probability expression parser."""

import unittest

from y0.dsl import C, D, P, Sum, Variable
from y0.parser import parse_causaleffect
from y0.parser.ce.grammar import sum_pe

U_1 = Variable("U_1")


class TestCausaleffectGrammar(unittest.TestCase):
    """Tests for parsing causaleffect probability expressions."""

    def test_sum(self):
        """Test the sum grammar."""
        for expr, s in [
            (
                Sum[U_1, C](P(D | U_1 | C) * P(C) * P(U_1)),
                r"\sum_{U_{1},C}P(D|U_{1},C)P(C)P(U_{1})",
            ),
        ]:
            with self.subTest(s=s):
                parse_result = sum_pe.parseString(s)
                result_expression = parse_result.asList()[0]
                self.assertIsInstance(result_expression, Sum)
                self.assertEqual(
                    expr,
                    result_expression,
                    msg=f"Mismatch\nExpected: {repr(expr)}\nActual:   {repr(result_expression)}",
                )

    def test_parse(self):
        """Test the high-level parser API."""
        for expr, s in [
            (
                Sum[U_1, C](P(D | U_1 | C) * P(C) * P(U_1)),
                r"\sum_{U_{1},C}P(D|U_{1},C)P(C)P(U_{1})",
            ),
        ]:
            with self.subTest(s=s):
                actual = parse_causaleffect(s)
                self.assertEqual(
                    expr, actual, msg=f"\nExpected: {expr}\nActual:   {actual}"
                )



# -*- coding: utf-8 -*-

"""Interface to the R causaleffect package via :mod:`rpy2`."""

from __future__ import annotations

import logging
from typing import Sequence, Union

from rpy2 import robjects

from .graph import CausalEffectGraph, NxMixedGraph
from .r_utils import uses_r
from .struct import VermaConstraint

logger = logging.getLogger(__name__)


@uses_r
def r_get_verma_constraints(
    graph: Union[NxMixedGraph, CausalEffectGraph]
) -> Sequence[VermaConstraint]:
    """Calculate the verma constraints on the graph using ``causaleffect``."""
    if isinstance(graph, NxMixedGraph):
        graph = graph.to_causaleffect()
    verma_constraints = robjects.r["verma.constraints"]
    return [VermaConstraint.from_element(row) for row in verma_constraints(graph)]



# -*- coding: utf-8 -*-

"""Implement Robin Evans' simplification algorithms.

.. seealso:: https://www.fields.utoronto.ca/programs/scientific/11-12/graphicmodels/Evans.pdf slides 34-43
"""

import itertools as itt
import logging
from typing import Iterable, Mapping, NamedTuple, Optional, Set, Tuple

import networkx as nx

from ..graph import DEFAULT_TAG

__all__ = [
    "simplify_latent_dag",
    "SimplifyResults",
    "remove_widow_latents",
    "transform_latents_with_parents",
    "remove_redundant_latents",
]

logger = logging.getLogger(__name__)

DEFAULT_SUFFIX = "_prime"


class SimplifyResults(NamedTuple):
    """Results from the simplification of a LV-DAG."""

    graph: nx.DiGraph
    widows: Set[str]
    redundant: Set[str]


def simplify_latent_dag(graph: nx.DiGraph, tag: Optional[str] = None):
    """Apply Robin Evans' three algorithms in succession."""
    if tag is None:
        tag = DEFAULT_TAG

    _ = transform_latents_with_parents(graph, tag=tag)
    _, widows = remove_widow_latents(graph, tag=tag)
    _, redundant = remove_redundant_latents(graph, tag=tag)

    return SimplifyResults(
        graph=graph,
        widows=widows,
        redundant=redundant,
    )


def iter_latents(graph: nx.DiGraph, *, tag: Optional[str] = None) -> Iterable[str]:
    """Iterate over nodes marked as latent.

    :param graph: A latent variable DAG
    :param tag: The tag for which variables are latent
    :yields: Nodes that are latent
    """
    if tag is None:
        tag = DEFAULT_TAG
    # should start with nodes the highest up (closest to source)
    for node in nx.topological_sort(graph):
        if graph.nodes[node][tag]:
            yield node


def remove_widow_latents(
    graph: nx.DiGraph, tag: Optional[str] = None
) -> Tuple[nx.DiGraph, Set[str]]:
    """Remove latents with no children (in-place).

    :param graph: A latent variable DAG
    :param tag: The tag for which variables are latent
    :returns: The graph, modified in place
    """
    remove = set(iter_widow_latents(graph, tag=tag))
    graph.remove_nodes_from(remove)
    return graph, remove


def iter_widow_latents(
    graph: nx.DiGraph, *, tag: Optional[str] = None
) -> Iterable[str]:
    """Iterate over latent variables with no children.

    :param graph: A latent variable DAG
    :param tag: The tag for which variables are latent
    :yields: Nodes with no children
    """
    for node in iter_latents(graph, tag=tag):
        if not graph.out_edges(node):
            yield node


def transform_latents_with_parents(
    graph: nx.DiGraph,
    tag: Optional[str] = None,
    suffix: Optional[str] = None,
) -> nx.DiGraph:
    """Transform latent variables with parents into latent variables with no parents.

    :param graph: A latent variable DAG
    :param tag: The tag for which variables are latent
    :param suffix: The suffix to postpend to transformed latent variables.
    :returns: The graph, modified in place
    """
    if tag is None:
        tag = DEFAULT_TAG
    if suffix is None:
        suffix = DEFAULT_SUFFIX
    for latent_node, parents, children in iter_middle_latents(graph, tag=tag):
        graph.remove_node(latent_node)
        graph.add_edges_from(itt.product(parents, children))

        new_node = f"{latent_node}{suffix}"
        graph.add_node(new_node, **{tag: True})
        for child in children:
            graph.add_edge(new_node, child)

    return graph


def _add_modified_latent(
    graph: nx.DiGraph,
    modified_latent: Mapping[str, Iterable[str]],
    *,
    tag: Optional[str] = None,
    suffix: Optional[str] = None,
) -> None:
    if tag is None:
        tag = DEFAULT_TAG
    if suffix is None:
        suffix = DEFAULT_SUFFIX

    for old_node, children in modified_latent.items():
        new_node = f"{old_node}{suffix}"
        graph.add_node(new_node, **{tag: True})
        for child in children:
            graph.add_edge(new_node, child)


def iter_middle_latents(
    graph: nx.DiGraph, *, tag: Optional[str] = None
) -> Iterable[Tuple[str, Set[str], Set[str]]]:
    """Iterate over latent nodes that have both parents and children (along with them).

    :param graph: A latent variable DAG
    :param tag: The tag for which variables are latent
    :yields: Nodes with both parents and children, along with the set of parents and set of children
    """
    for node in iter_latents(graph, tag=tag):
        parents = set(graph.predecessors(node))
        if 0 == len(parents):
            continue
        children = set(graph.successors(node))
        if 0 == len(children):
            continue
        yield node, parents, children


def remove_redundant_latents(
    graph: nx.DiGraph, tag: Optional[str] = None
) -> Tuple[nx.DiGraph, Set[str]]:
    """Remove redundant latent variables.

    :param graph: A latent variable DAG
    :param tag: The tag for which variables are latent
    :returns: The graph, modified in place
    """
    remove = set(_iter_redundant_latents(graph, tag=tag))
    graph.remove_nodes_from(remove)
    return graph, remove


def _iter_redundant_latents(graph, *, tag: Optional[str] = None) -> Iterable[str]:
    latents: Mapping[str, Set[str]] = {
        node: set(graph.successors(node)) for node in iter_latents(graph, tag=tag)
    }
    for (left, left_children), (right, right_children) in itt.product(
        latents.items(), repeat=2
    ):
        if left_children == right_children and left > right:
            # if children are the same, keep the lower sort order node
            yield left
        elif left_children < right_children:
            # if left's children are a proper subset of right's children, we don't need left
            yield left



# -*- coding: utf-8 -*-
# type: ignore

"""Examples from CausalFusion."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Sequence

import networkx as nx
import pandas as pd

from .dsl import P, Q, Sum, Variable, X, Y, Z1, Z2, Z3, Z4, Z5
from .graph import NxMixedGraph
from .resources import ASIA_PATH
from .struct import DSeparationJudgement, VermaConstraint


@dataclass
class Example:
    """An example graph packaged with certain pre-calculated data structures."""

    name: str
    reference: str
    graph: NxMixedGraph
    verma_constraints: Optional[Sequence[VermaConstraint]] = None
    conditional_independencies: Optional[Sequence[DSeparationJudgement]] = None
    data: Optional[pd.DataFrame] = None


u_2 = Variable("u_2")
u_3 = Variable("u_3")

#: Treatment: X
#: Outcome: Y
#: Adjusted: N/A
backdoor = NxMixedGraph.from_edges(
    directed=[
        ("Z", "X"),
        ("Z", "Y"),
        ("X", "Y"),
    ]
)

backdoor_example = Example(
    name="Backdoor",
    reference='J. Pearl. 2009. "Causality: Models, Reasoning and Inference.'
    ' 2nd ed." Cambridge University Press, p. 178.',
    graph=backdoor,
)

#: Treatment: X
#: Outcome: Y
#: Adjusted: N/A
frontdoor = NxMixedGraph.from_edges(
    directed=[
        ("X", "Z"),
        ("Z", "Y"),
    ],
    undirected=[
        ("X", "Y"),
    ],
)
frontdoor_example = Example(
    name="Frontdoor",
    reference='J. Pearl. 2009. "Causality: Models, Reasoning and Inference.'
    ' 2nd ed." Cambridge University Press, p. 81.',
    graph=frontdoor,
)

#: Treatment: X
#: Outcome: Y
instrumental_variable = NxMixedGraph.from_edges(
    directed=[
        ("Z", "X"),
        ("X", "Y"),
    ],
    undirected=[
        ("X", "Y"),
    ],
)
instrumental_variable_example = Example(
    name="Instrument Variable",
    reference='J. Pearl. 2009. "Causality: Models, Reasoning and Inference.'
    ' 2nd ed." Cambridge University Press, p. 153.',
    graph=instrumental_variable,
)

#: Treatment: X
#: Outcome: Y
napkin = NxMixedGraph.from_edges(
    directed=[
        ("Z2", "Z1"),
        ("Z1", "X"),
        ("X", "Y"),
    ],
    undirected=[
        ("Z2", "X"),
        ("Z2", "Y"),
    ],
)
napkin_example = Example(
    name="Napkin",
    reference='J. Pearl and D. Mackenzie. 2018. "The Book of Why: The New Science of Cause and Effect."'
    " Basic Books, p. 240.",
    graph=napkin,
    verma_constraints=[
        VermaConstraint(
            lhs_cfactor=Q[X, Y](Z1, X, Y) / Sum[Y](Q[X, Y](Z1, X, Y)),
            lhs_expr=(
                Sum[Z2](P(Y | (Z1, Z2, X)) * P(X | (Z2, Z1)) * P(Z2))
                / Sum[Z2, Y](P(Y | (Z2, Z1, X)) * P(X | (Z2, Z1)) * P(Z2))
            ),
            rhs_cfactor=Q[Y](X, Y),
            rhs_expr=Sum[u_2, X](P(Y | u_2 | X) * P(X) * P(u_2)),
            variables=(Z1,),
        ),
    ],
)

#: Treatment: X
#: Outcome: Y
#: Reference:
m_graph = NxMixedGraph.from_edges(
    directed=[
        ("X", "Y"),
    ],
    undirected=[
        ("X", "Z"),
        ("Y", "Z"),
    ],
)
m_graph_example = Example(
    name="M-Graph",
    reference='S. Greenland, J. Pearl, and J.M. Robins. 1999. "Causal Diagrams for Epidemiologic Research."'
    " Epidemiology Journal, Volume 10, No. 10, pp. 37-48, 1999.",
    graph=m_graph,
)

#: Treatment: X
#: Outcome: Y
identifiability_1 = NxMixedGraph.from_edges(
    directed=[
        ("Z1", "Z2"),
        ("Z1", "Z3"),
        ("Z2", "X"),
        ("Z3", "X"),
        ("Z4", "X"),
        ("Z4", "Z5"),
        ("Z3", "Y"),
        ("X", "Y"),
        ("Z3", "Y"),
    ],
)
identifiability_1_example = Example(
    name="Identifiability 1",
    reference='J. Pearl. 2009. "Causality: Models, Reasoning and Inference.'
    ' 2nd ed." Cambridge University Press, p. 80.',
    graph=identifiability_1,
    conditional_independencies=(
        DSeparationJudgement.create("X", "Z1", ["Z2", "Z3"]),
        DSeparationJudgement.create("X", "Z5", ["Z4"]),
        DSeparationJudgement.create("Y", "Z1", ["X", "Z3", "Z4"]),
        DSeparationJudgement.create("Y", "Z2", ["X", "Z1", "Z3"]),
        DSeparationJudgement.create("Y", "Z4", ["X", "Z3", "Z5"]),
        DSeparationJudgement.create("Z1", "Z4"),
        DSeparationJudgement.create("Z1", "Z5"),
        DSeparationJudgement.create("Z2", "Z3", ["Z1"]),
        DSeparationJudgement.create("Z2", "Z4"),
        DSeparationJudgement.create("Z2", "Z5"),
        DSeparationJudgement.create("Z3", "Z5"),
        DSeparationJudgement.create("Y", "Z5", ["X", "Z3"]),
        DSeparationJudgement.create("Z3", "Z4"),
    ),
)

#: Treatment: X
#: Outcome: Y
identifiability_2 = NxMixedGraph.from_edges(
    directed=[
        ("Z1", "Z2"),
        ("Z1", "Z3"),
        ("Z2", "X"),
        ("Z3", "X"),
        ("X", "W0"),
        ("W0", "Y"),
        ("Z4", "Z3"),
        ("Z4", "Z5"),
        ("Z5", "Y"),
        ("X", "W1"),
        ("W1", "W2"),
        ("W2", "Y"),
        ("Z4", "Z3"),
        ("Z3", "Y"),
    ],
    undirected=[
        ("Z1", "X"),
        ("Z2", "Z3"),
        ("Z3", "Z5"),
        ("Z4", "Y"),
    ],
)

identifiability_2_example = Example(
    name="Identifiability 2",
    reference="E. Bareinboim modification of Identifiability 1.",
    graph=identifiability_2,
    verma_constraints=[
        VermaConstraint(
            rhs_cfactor=Q[Z5](Z4, Z5),
            rhs_expr=Sum[u_3, Z4](P(Z5 | (u_3, Z4)) * P(Z4) * P(u_3)),
            lhs_cfactor=Sum[Z3](Q[Z3, Z5](Z1, Z4, Z3, Z5)),
            lhs_expr=Sum[Z3](P(Z5 | (Z1, Z2, Z3, Z4)) * P(Z3 | (Z1, Z2, Z4))),
            variables=(Z1,),
        ),
        VermaConstraint(
            rhs_cfactor=Q[Z5](Z4, Z5),
            rhs_expr=Sum[u_3, Z4](P(Z5 | (u_3, Z4)) * P(Z4) * P(u_3)),
            lhs_cfactor=(
                Q[Z2, Z5](Z1, Z4, Z2, Z5) / Sum[Z5](Q[Z2, Z5](Z1, Z4, Z2, Z5))
            ),
            lhs_expr=(
                Sum[Z3](
                    P(Z5 | (Z1, Z2, Z3, Z4)) * P(Z3 | (Z1, Z4, Z2)) * P(Z2 | (Z1, Z4))
                )
                / Sum[Z3, Z5](
                    P(Z5 | (Z1, Z4, Z2, Z3)) * P(Z3 | (Z1, Z4, Z2)) * P(Z2 | (Z1, Z4))
                )
            ),
            variables=(Z1, Z2),
        ),
    ],
    conditional_independencies=[
        DSeparationJudgement.create("W0", "W1", ["X"]),
        DSeparationJudgement.create("W0", "W2", ["X"]),
        DSeparationJudgement.create("W0", "Z1", ["X"]),
        DSeparationJudgement.create("W0", "Z2", ["X"]),
        DSeparationJudgement.create("W0", "Z3", ["X"]),
        DSeparationJudgement.create("W0", "Z4", ["X"]),
        DSeparationJudgement.create("W0", "Z5", ["X"]),
        DSeparationJudgement.create("W1", "Y", ["W0", "W2", "Z3", "Z4", "Z5"]),
        DSeparationJudgement.create("W1", "Z1", ["X"]),
        DSeparationJudgement.create("W1", "Z2", ["X"]),
        DSeparationJudgement.create("W1", "Z3", ["X"]),
        DSeparationJudgement.create("W1", "Z4", ["X"]),
        DSeparationJudgement.create("W1", "Z5", ["X"]),
        DSeparationJudgement.create("W2", "X", ["W1"]),
        DSeparationJudgement.create("W2", "Z1", ["W1"]),
        DSeparationJudgement.create("W2", "Z2", ["W1"]),
        DSeparationJudgement.create("W2", "Z3", ["W1"]),
        DSeparationJudgement.create("W2", "Z4", ["W1"]),
        DSeparationJudgement.create("W2", "Z5", ["W1"]),
        DSeparationJudgement.create("X", "Y", ["W0", "W2", "Z3", "Z4", "Z5"]),
        DSeparationJudgement.create("X", "Z4", ["Z1", "Z2", "Z3"]),
        DSeparationJudgement.create("X", "Z5", ["Z1", "Z2", "Z3"]),
        DSeparationJudgement.create("Y", "Z1", ["W0", "W2", "Z3", "Z4", "Z5"]),
        DSeparationJudgement.create("Y", "Z2", ["W0", "W2", "Z3", "Z4", "Z5"]),
        DSeparationJudgement.create("Z1", "Z4"),
        DSeparationJudgement.create("Z1", "Z5"),
        DSeparationJudgement.create("Z2", "Z4"),
        DSeparationJudgement.create("Z2", "Z5"),
    ],
)

#: The Identifiability 3 example
#: Treatment: X
#: Outcome: Y
#: Reference: J. Pearl. 2009. "Causality: Models, Reasoning and Inference. 2nd ed." Cambridge University Press, p. 92.
identifiability_3 = NxMixedGraph.from_edges(
    directed=[
        ("Z2", "X"),
        ("Z2", "Z1"),
        ("Z2", "Z3"),
        ("X", "Z1"),
        ("Z3", "Y"),
        ("Z1", "Y"),
    ],
    undirected=[
        ("Z2", "X"),
        ("Z2", "Y"),
        ("X", "Z3"),
        ("X", "Y"),
    ],
)

#: The Identifiability 4 example
#: Treatment: X
#: Outcome: Y
#: Reference: J. Pearl. 2009. "Causality: Models, Reasoning and Inference. 2nd ed." Cambridge University Press, p. 92.
identifiability_4 = NxMixedGraph.from_edges(
    directed=[
        ("X", "Z1"),
        ("X", "Y"),
        ("Z1", "Z2"),
        ("Z1", "Y"),
        ("Z2", "Y"),
    ],
    undirected=[
        ("X", "Z2"),
        ("Z1", "Y"),
    ],
)

#: The Identifiability 5 example
#: Treatment: X1, X2
#: Outcome: Y
#: Reference: J. Pearl. 2009. "Causality: Models, Reasoning and Inference. 2nd ed." Cambridge University Press, p. 119.
identifiability_5 = NxMixedGraph.from_edges(
    directed=[
        ("X1", "Z"),
        ("X1", "Y"),
        ("X1", "X2"),
        ("Z", "X2"),
        ("X2", "Y"),
    ],
    undirected=[
        ("X1", "Z"),
        ("Z", "Y"),
    ],
)

#: The Identifiability 6 example
#: Treatment: X1, X2
#: Outcome: Y
#: Reference: J. Pearl. 2009. "Causality: Models, Reasoning and Inference. 2nd ed." Cambridge University Press, p. 125.
identifiability_6 = NxMixedGraph.from_edges(
    directed=[
        ("Z1", "X1"),
        ("X1", "X2"),
        ("X2", "Y"),
        ("Z2", "Y"),
    ],
    undirected=[
        ("Z1", "Z2"),
        ("Z1", "X2"),
        ("Z2", "X2"),
    ],
)

#: The Identifiability 7 example
#: Treatment: X
#: Outcome: Y
#: Reference: J. Tian. 2002. "Studies in Causal Reasoning and Learning." p. 90.
identifiability_7 = NxMixedGraph.from_edges(
    directed=[
        ("W1", "W2"),
        ("W3", "W4"),
        ("W2", "X"),
        ("W4", "X"),
        ("X", "Y"),
    ],
    undirected=[
        ("W1", "X"),
        ("W1", "Y"),
        ("W1", "W3"),
        ("W3", "W2"),
        ("W3", "W5"),
        ("W5", "W4"),
    ],
)

# TODO Recoverability 1/2 - what is the S node?
# TODO Transportability 1/2 - what are the box nodes?
# TODO g-Identifiability examples
# TODO g-Transportability examples


#: The Verma 1 example
#: Treatment: V3
#: Outcome: V4
#: Reference: T. Verma and J. Pearl. 1990. "Equivalence and Synthesis of Causal Models." In P. Bonissone et al., eds.,
#: Proceedings of the 6th Conference on Uncertainty in Artificial Intelligence. Cambridge, MA: AUAI Press, p. 257.
verma_1 = NxMixedGraph.from_edges(
    directed=[
        ("V1", "V2"),
        ("V2", "V3"),
        ("V3", "V4"),
    ],
    undirected=[
        ("V2", "V4"),
    ],
)

#: The Verma 2 example
#: Treatment: V1
#: Outcome: V5
#: Reference: J. Tian. 2002. "Studies in Causal Reasoning and Learning." p. 70.
verma_2 = NxMixedGraph.from_edges(
    directed=[
        ("V1", "V2"),
        ("V2", "V3"),
        ("V3", "V4"),
        ("V4", "V5"),
    ],
    undirected=[
        ("V1", "V3"),
        ("V2", "V4"),
        ("V3", "V5"),
    ],
)

#: The Verma 3 example
#: Treatment: V1
#: Outcome: V5
#: Reference: J. Tian. 2002. "Studies in Causal Reasoning and Learning." p. 59.
verma_3 = NxMixedGraph.from_edges(
    directed=[
        ("V1", "V2"),
        ("V2", "V3"),
        ("V3", "V4"),
        ("V4", "V5"),
    ],
    undirected=[
        ("V1", "V5"),
        ("V1", "V3"),
        ("V2", "V4"),
    ],
)

#: The Verma 4 example
#: Treatment: V1
#: Outcome: V5
#: Reference: E. Bareinboim modification of Verma 2.
verma_4 = NxMixedGraph.from_edges(
    directed=[
        ("V1", "V2"),
        ("V2", "V3"),
        ("V3", "V4"),
        ("V4", "V5"),
    ],
    undirected=[
        ("V1", "V5"),
        ("V1", "V3"),
        ("V2", "V4"),
        ("V3", "V5"),
    ],
)

#: The Verma 5 example
#: Treatment: V1
#: Outcome: V5
#: Reference: E. Bareinboim modification of Verma 2.
verma_5 = NxMixedGraph.from_edges(
    directed=[
        ("V1", "V2"),
        ("V2", "V3"),
        ("V3", "V4"),
        ("V4", "V5"),
        ("V5", "V6"),
    ],
    undirected=[
        ("V0", "V1"),
        ("V0", "V6"),
        ("V1", "V5"),
        ("V1", "V3"),
        ("V2", "V4"),
    ],
)

#: The z-Identifiability 1 example
#: Treatment: X
#: Outcome: Y
#: Z*: Z
#: Reference: E. Bareinboim and J. Pearl. 2012. "Causal Inference by Surrogate Experiments: z-Identifiability." In
#: Nando de Freitas and K. Murphy., eds., Proceedings of the 28th Conference on Uncertainty in Artificial Intelligence.
#: Corvallis, OR: AUAI Press, p. 114.
z_identifiability_1 = NxMixedGraph.from_edges(
    directed=[
        ("Z", "X"),
        ("X", "Y"),
    ],
    undirected=[
        ("Z", "X"),
        ("Z", "Y"),
    ],
)

#: The z-Identifiability 2 example
#: Treatment: X
#: Outcome: Y
#: Z*: Z
#: Reference: E. Bareinboim and J. Pearl. 2012. "Causal Inference by Surrogate Experiments: z-Identifiability." In
#: Nando de Freitas and K. Murphy., eds., Proceedings of the 28th Conference on Uncertainty in Artificial Intelligence.
#: Corvallis, OR: AUAI Press, p. 114.
z_identifiability_2 = NxMixedGraph.from_edges(
    directed=[
        ("Z", "X"),
        ("X", "Y"),
    ],
    undirected=[
        ("X", "Y"),
        ("Z", "Y"),
    ],
)

#: The z-Identifiability 3 example
#: Treatment: X
#: Outcome: Y
#: Z*: Z
#: Reference: E. Bareinboim and J. Pearl. 2012. "Causal Inference by Surrogate Experiments: z-Identifiability." In
#: Nando de Freitas and K. Murphy., eds., Proceedings of the 28th Conference on Uncertainty in Artificial Intelligence.
#: Corvallis, OR: AUAI Press, p. 114.
z_identifiability_3 = NxMixedGraph.from_edges(
    directed=[
        ("Z", "Y"),
        ("X", "Y"),
    ],
    undirected=[
        ("X", "Z"),
        ("Z", "Y"),
    ],
)

#: The Identifiability (Linear) 1 example
#: Treatment: X
#: Outcome: Y
#: Reference: J. Pearl. 2009. "Causality: Models, Reasoning and Inference. 2nd ed." Cambridge University Press, p. 153.
identifiability_linear_1 = NxMixedGraph.from_edges(
    directed=[
        ("X", "Z"),
        ("X", "W"),
        ("W", "Y"),
        ("Z", "Y"),
    ],
    undirected=[
        ("X", "Z"),
        ("W", "Y"),
    ],
)

d_separation_example = Example(
    name="D-separation example",
    reference="http://web.mit.edu/jmn/www/6.034/d-separation.pdf",
    graph=NxMixedGraph.from_edges(
        directed=[
            ("AA", "C"),
            ("B", "C"),
            ("C", "D"),
            ("C", "E"),
            ("D", "F"),
            ("F", "G"),
        ],
    ),
    conditional_independencies=[
        DSeparationJudgement.create("AA", "B"),
        DSeparationJudgement.create("AA", "D", ["C"]),
        DSeparationJudgement.create("AA", "E", ["C"]),
        DSeparationJudgement.create("AA", "F", ["C"]),
        DSeparationJudgement.create("AA", "G", ["C"]),
        DSeparationJudgement.create("B", "D", ["C"]),
        DSeparationJudgement.create("B", "E", ["C"]),
        DSeparationJudgement.create("B", "F", ["C"]),
        DSeparationJudgement.create("B", "G", ["C"]),
        DSeparationJudgement.create("C", "F", ["D"]),
        DSeparationJudgement.create("C", "G", ["D"]),
        DSeparationJudgement.create("D", "E", ["C"]),
        DSeparationJudgement.create("D", "G", ["F"]),
        DSeparationJudgement.create("E", "F", ["C"]),
        DSeparationJudgement.create("E", "G", ["C"]),
    ],
)

asia_example = Example(
    name="Asia dataset",
    reference="https://www.bnlearn.com/documentation/man/asia.html",
    graph=NxMixedGraph.from_edges(
        directed=[
            ("Asia", "Tub"),
            ("Smoke", "Lung"),
            ("Smoke", "Bronc"),
            ("Tub", "Either"),
            ("Lung", "Either"),
            ("Either", "Xray"),
            ("Either", "Dysp"),
            ("Bronc", "Dysp"),
        ],
    ),
    data=pd.read_csv(ASIA_PATH).replace({"yes": 1, "no": -1}),
)

examples = [v for name, v in locals().items() if name.endswith("_example")]

#: The IGF directed graph example from Sara
igf_graph = nx.DiGraph(
    [
        ("EGF", "SOS"),
        ("EGF", "PI3K"),
        ("IGF", "SOS"),
        ("IGF", "PI3K"),
        ("SOS", "Ras"),
        ("Ras", "PI3K"),
        ("Ras", "Raf"),
        ("PI3K", "Akt"),
        ("Akt", "Raf"),
        ("Raf", "Mek"),
        ("Mek", "Erk"),
    ]
)



# -*- coding: utf-8 -*-

"""Utilities."""

import math
from itertools import chain, combinations
from typing import Any, Collection, Iterable, Mapping, Optional, TypeVar

from tqdm import tqdm

__all__ = [
    "powerset",
]

X = TypeVar("X")


def powerset(
    iterable: Iterable[X],
    start: int = 0,
    stop: Optional[int] = None,
    *,
    reverse: bool = False,
    use_tqdm: Optional[bool] = False,
    tqdm_kwargs: Optional[Mapping[str, Any]] = None,
) -> Iterable[Collection[X]]:
    """Get successively longer combinations of the source.

    :param iterable: List to get combinations from
    :param start: smallest combination to get (default 0)
    :param stop: Largest combination to get (None means length of the list and is the default)
    :param reverse: Should the bigger powersets be returned first?
    :param use_tqdm: Should a progress bar be shown
    :param tqdm_kwargs: Options for tqdm
    :return: Iterator of powerset of values.

    .. seealso: :func:`more_iterools.powerset` for a non-constrainable implementation
    """
    s = list(iterable)
    n = len(s)
    if stop is None:
        stop = n + 1

    if reverse:
        rv = chain.from_iterable(combinations(s, n - r) for r in range(start, stop))
    else:
        rv = chain.from_iterable(combinations(s, r) for r in range(start, stop))

    if use_tqdm:
        total = sum(math.comb(n, k) for k in range(start, stop))
        rv = tqdm(rv, disable=use_tqdm, total=total, **(tqdm_kwargs or {}))

    return rv



# -*- coding: utf-8 -*-

"""General utilities for :mod:`rpy2`."""

import logging
from functools import lru_cache, wraps
from typing import Iterable, Tuple

from rpy2.robjects.packages import importr, isinstalled
from rpy2.robjects.vectors import StrVector

from .dsl import Variable

__all__ = [
    "uses_r",
]

logger = logging.getLogger(__name__)

CAUSALEFFECT = "causaleffect"
IGRAPH = "igraph"
R_REQUIREMENTS = [
    CAUSALEFFECT,
    IGRAPH,
]


def prepare_renv(requirements: Iterable[str]) -> None:
    """Ensure the given R packages are installed.

    :param requirements: A list of R packages to ensure are installed

    .. seealso:: https://rpy2.github.io/doc/v3.4.x/html/introduction.html#installing-packages
    """
    # import R's utility package
    utils = importr("utils")

    # select a mirror for R packages
    utils.chooseCRANmirror(ind=1)  # select the first mirror in the list

    uninstalled_requirements = [
        requirement for requirement in requirements if not isinstalled(requirement)
    ]
    if uninstalled_requirements:
        logger.warning("installing R packages: %s", uninstalled_requirements)
        utils.install_packages(StrVector(uninstalled_requirements))

    for requirement in requirements:
        importr(requirement)


@lru_cache(maxsize=1)
def prepare_default_renv() -> bool:
    """Prepare the default R environment."""
    prepare_renv(R_REQUIREMENTS)
    return True


def uses_r(f):
    """Decorate functions that use R."""

    @wraps(f)
    def _wrapped(*args, **kwargs):
        prepare_default_renv()
        return f(*args, **kwargs)

    return _wrapped


def _parse_vars(element) -> Tuple[Variable, ...]:
    _vars = element.rx("vars")
    return tuple(Variable(name) for name in sorted(_vars[0]))


def _extract(element, key):
    return element.rx(key)[0][0]



# -*- coding: utf-8 -*-

"""An internal domain-specific language for probability expressions."""

from __future__ import annotations

import functools
import itertools as itt
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from operator import attrgetter
from typing import Callable, Iterable, Optional, Sequence, Set, Tuple, TypeVar, Union

__all__ = [
    "Variable",
    "Intervention",
    "CounterfactualVariable",
    "Distribution",
    "P",
    "Probability",
    "Sum",
    "Product",
    "Fraction",
    "Expression",
    "One",
    "Q",
    "QFactor",
    "A",
    "B",
    "C",
    "D",
    "R",
    "S",
    "T",
    "W",
    "X",
    "Y",
    "Z",
    "V1",
    "V2",
    "V3",
    "V4",
    "V5",
    "V6",
    "Z1",
    "Z2",
    "Z3",
    "Z4",
    "Z5",
    "Z6",
    # Helpers
    "ensure_ordering",
]

X = TypeVar("X")
XSeq = Union[X, Sequence[X]]


def _upgrade_variables(variables: XSeq[Variable]) -> Tuple[Variable, ...]:
    return (variables,) if isinstance(variables, Variable) else tuple(variables)


def _to_interventions(variables: Sequence[Variable]) -> Tuple[Intervention, ...]:
    return tuple(
        variable
        if isinstance(variable, Intervention)
        else Intervention(name=variable.name, star=False)
        for variable in variables
    )


class _Mathable(ABC):
    @abstractmethod
    def to_text(self) -> str:
        """Output this DSL object in the internal string format."""

    @abstractmethod
    def to_latex(self) -> str:
        """Output this DSL object in the LaTeX string format."""

    def _repr_latex_(self) -> str:  # hack for auto-display of latex in jupyter notebook
        return f"${self.to_latex()}$"

    def __str__(self) -> str:
        return self.to_text()

    @abstractmethod
    def _iter_variables(self) -> Iterable[Variable]:
        """Iterate over variables."""

    def get_variables(self) -> Set[Variable]:
        """Get the set of variables used in this expression."""
        return set(self._iter_variables())


@dataclass(frozen=True)
class Variable(_Mathable):
    """A variable, typically with a single letter."""

    #: The name of the variable
    name: str

    def __post_init__(self):
        if self.name in {"P", "Q"}:
            raise ValueError(f"trust me, {self.name} is a bad variable name.")

    @classmethod
    def norm(cls, name: Union[str, Variable]) -> Variable:
        """Automatically upgrade a string to a variable."""
        if isinstance(name, str):
            return Variable(name)
        elif isinstance(name, Variable):
            return name
        else:
            raise TypeError

    def to_text(self) -> str:
        """Output this variable in the internal string format."""
        return self.name

    def to_latex(self) -> str:
        """Output this variable in the LaTeX string format."""
        return self.to_text()

    def intervene(self, variables: XSeq[Variable]) -> CounterfactualVariable:
        """Intervene on this variable with the given variable(s).

        :param variables: The variable(s) used to extend this variable as it is changed to a
            counterfactual variable
        :returns: A new counterfactual variable over this variable with the given intervention(s).

        .. note:: This function can be accessed with the matmult @ operator.
        """
        return CounterfactualVariable(
            name=self.name,
            interventions=_to_interventions(_upgrade_variables(variables)),
        )

    def __matmul__(self, variables: XSeq[Variable]) -> CounterfactualVariable:
        return self.intervene(variables)

    def given(self, parents: Union[XSeq[Variable], Distribution]) -> Distribution:
        """Create a distribution in which this variable is conditioned on the given variable(s).

        The new distribution is a Markov Kernel.

        :param parents: A variable or list of variables to include as conditions in the new conditional distribution
        :returns: A new conditional probability distribution
        :raises TypeError: If a distribution is given as the parents that contains conditionals

        .. note:: This function can be accessed with the or | operator.
        """
        if not isinstance(parents, Distribution):
            return Distribution(
                children=(self,),
                parents=_upgrade_variables(parents),
            )
        elif parents.is_conditioned():
            raise TypeError("can not be given a distribution that has conditionals")
        else:
            # The parents variable is actually a Distribution instance with no parents,
            #  so its children become the parents for the new Markov Kernel distribution
            return Distribution(
                children=(self,),
                parents=parents.children,  # don't think about this too hard
            )

    def __or__(self, parents: XSeq[Variable]) -> Distribution:
        return self.given(parents)

    def joint(self, children: XSeq[Variable]) -> Distribution:
        """Create a joint distribution between this variable and the given variable(s).

        :param children: The variable(s) for use with this variable in a joint distribution
        :returns: A new joint distribution over this variable and the given variables.

        .. note:: This function can be accessed with the and & operator.
        """
        return Distribution(
            children=(self, *_upgrade_variables(children)),
        )

    def __and__(self, children: XSeq[Variable]) -> Distribution:
        return self.joint(children)

    def invert(self) -> Intervention:
        """Create an :class:`Intervention` variable that is different from what was observed (with a star)."""
        return Intervention(name=self.name, star=True)

    def __invert__(self) -> Intervention:
        return self.invert()

    def __neg__(self) -> Intervention:
        return Intervention(name=self.name, star=False)

    @classmethod
    def __class_getitem__(cls, item) -> Variable:
        return Variable(item)

    def _iter_variables(self) -> Iterable[Variable]:
        """Get a set containing this variable."""
        yield self


@dataclass(frozen=True)
class Intervention(Variable):
    """An intervention variable.

    An intervention variable is usually used as a subscript in a :class:`CounterfactualVariable`.
    """

    #: The name of the intervention
    name: str
    #: If true, indicates this intervention represents a value different from what was observed
    star: bool = False

    def to_text(self) -> str:
        """Output this intervention variable in the internal string format."""
        return f"{self.name}*" if self.star else self.name

    def to_latex(self) -> str:
        """Output this intervention variable in the LaTeX string format."""
        return f"{self.name}^*" if self.star else self.name

    def invert(self) -> Intervention:
        """Create an :class:`Intervention` variable that is different from what was observed (with a star)."""
        return Intervention(name=self.name, star=not self.star)


@dataclass(frozen=True)
class CounterfactualVariable(Variable):
    """A counterfactual variable.

    Counterfactual variables are like normal variables, but can have a list of interventions.
    Each intervention is either the same as what was observed (no star) or different from what
    was observed (star).
    """

    #: The name of the counterfactual variable
    name: str
    #: The interventions on the variable. Should be non-empty
    interventions: Tuple[Intervention, ...]

    def __post_init__(self):
        if not self.interventions:
            raise ValueError("should give at least one intervention")
        for intervention in self.interventions:
            if not isinstance(intervention, Intervention):
                raise TypeError(
                    f"only Intervention instances are allowed."
                    f" Got: ({intervention.__class__.__name__}) {intervention}",
                )

    def to_text(self) -> str:
        """Output this counterfactual variable in the internal string format."""
        intervention_latex = _list_to_text(self.interventions)
        return f"{self.name}_{{{intervention_latex}}}"

    def to_latex(self) -> str:
        """Output this counterfactual variable in the LaTeX string format."""
        intervention_latex = _list_to_latex(self.interventions)
        return f"{self.name}_{{{intervention_latex}}}"

    def intervene(self, variables: XSeq[Variable]) -> CounterfactualVariable:
        """Intervene on this counterfactual variable with the given variable(s).

        :param variables: The variable(s) used to extend this counterfactual variable's
            current interventions. Automatically converts variables to interventions.
        :returns: A new counterfactual variable with both this counterfactual variable's interventions
            and the given intervention(s)

        .. note:: This function can be accessed with the matmult @ operator.
        """
        _variables = _upgrade_variables(variables)
        _interventions = _to_interventions(_variables)
        self._raise_for_overlapping_interventions(_interventions)
        return CounterfactualVariable(
            name=self.name,
            interventions=(*self.interventions, *_interventions),
        )

    def _raise_for_overlapping_interventions(
        self, interventions: Iterable[Intervention]
    ) -> None:
        """Raise an error if any of the given variables are already listed in interventions in this counterfactual.

        :param interventions: Interventions to check for overlap
        :raises ValueError: If there are overlapping variables given.
        """
        overlaps = {
            new
            for old, new in itt.product(self.interventions, interventions)
            if old.name == new.name
        }
        if overlaps:
            raise ValueError(
                f"Overlapping interventions in new interventions: {overlaps}"
            )

    def invert(self) -> Intervention:
        """Raise an error, since counterfactuals can't be inverted the same as normal variables or interventions."""
        raise NotImplementedError

    def _iter_variables(self) -> Iterable[Variable]:
        """Get the union of this variable and its interventions."""
        yield from super()._iter_variables()
        for intervention in self.interventions:
            yield from intervention._iter_variables()


@dataclass(frozen=True)
class Distribution(_Mathable):
    """A general distribution over several child variables, conditioned by several parents."""

    children: Tuple[Variable, ...]
    parents: Tuple[Variable, ...] = field(default_factory=tuple)

    def __post_init__(self):
        if isinstance(self.children, (list, Variable)):
            raise TypeError(f"children of wrong type: {type(self.children)}")
        if isinstance(self.parents, (list, Variable)):
            raise TypeError
        if not self.children:
            raise ValueError("distribution must have at least one child")

    def to_text(self) -> str:
        """Output this distribution in the internal string format."""
        children = _list_to_text(self.children)
        if self.parents:
            parents = _list_to_text(self.parents)
            return f"{children}|{parents}"
        else:
            return children

    def to_latex(self) -> str:
        """Output this distribution in the LaTeX string format."""
        children = _list_to_latex(self.children)
        if self.parents:
            parents = _list_to_latex(self.parents)
            return f"{children}|{parents}"
        else:
            return children

    def is_conditioned(self) -> bool:
        """Return if this distribution is conditioned."""
        return 0 < len(self.parents)

    def is_markov_kernel(self) -> bool:
        """Return if this distribution a markov kernel -> one child variable and one or more conditionals."""
        return len(self.children) == 1

    def uncondition(self) -> Distribution:
        """Return a new distribution that is not conditioned on the parents."""
        return Distribution(
            children=(*self.children, *self.parents),
        )

    def joint(self, children: XSeq[Variable]) -> Distribution:
        """Create a new distribution including the given child variables.

        :param children: The variable(s) with which this distribution's children are extended
        :returns: A new distribution.

        .. note:: This function can be accessed with the and & operator.
        """
        return Distribution(
            children=(*self.children, *_upgrade_variables(children)),
            parents=self.parents,
        )

    def __and__(self, children: XSeq[Variable]) -> Distribution:
        return self.joint(children)

    def given(self, parents: Union[XSeq[Variable], Distribution]) -> Distribution:
        """Create a new mixed distribution additionally conditioned on the given parent variables.

        :param parents: The variable(s) with which this distribution's parents are extended
        :returns: A new distribution
        :raises TypeError: If a distribution is given as the parents that contains conditionals

        .. note:: This function can be accessed with the or | operator.
        """
        if not isinstance(parents, Distribution):
            return Distribution(
                children=self.children,
                parents=(*self.parents, *_upgrade_variables(parents)),
            )
        elif parents.is_conditioned():
            raise TypeError("can not be given a distribution that has conditionals")
        else:
            # The parents variable is actually a Distribution instance with no parents,
            #  so its children get appended as parents for the new mixed distribution
            return Distribution(
                children=self.children,
                parents=(
                    *self.parents,
                    *parents.children,
                ),  # don't think about this too hard
            )

    def __or__(self, parents: XSeq[Variable]) -> Distribution:
        return self.given(parents)

    def _iter_variables(self) -> Iterable[Variable]:
        """Get the set of variables used in this distribution."""
        for variable in itt.chain(self.children, self.parents):
            yield from variable._iter_variables()


class Expression(_Mathable, ABC):
    """The abstract class representing all expressions."""

    @abstractmethod
    def __mul__(self, other):
        pass

    @abstractmethod
    def __truediv__(self, other):
        pass


@dataclass(frozen=True)
class Probability(Expression):
    """The probability over a distribution."""

    #: The distribution over which the probability is expressed
    distribution: Distribution

    def to_text(self) -> str:
        """Output this probability in the internal string format."""
        return f"P({self.distribution.to_text()})"

    def to_latex(self) -> str:
        """Output this probability in the LaTeX string format."""
        return f"P({self.distribution.to_latex()})"

    def __repr__(self):
        return f"P({repr(self.distribution)})"

    def __mul__(self, other: Expression) -> Expression:
        if isinstance(other, Product):
            return Product((self, *other.expressions))
        elif isinstance(other, Fraction):
            return Fraction(self * other.numerator, other.denominator)
        else:
            return Product((self, other))

    def __truediv__(self, expression: Expression) -> Fraction:
        return Fraction(self, expression)

    def uncondition(self) -> Probability:
        """Return a new probability where the underlying distribution is no longer conditioned by the parents.

        :returns: A new probability over a distribution over the children and parents of the previous distribution

        >>> from y0.dsl import P, A, B
        >>> P(A | B).uncondition() == P(A, B)
        """
        return Probability(self.distribution.uncondition())

    def marginalize(self, ranges: XSeq[Variable]) -> Fraction:
        """Return this probability distribution, marginalized by the given variables.

        :param ranges: A variable or list of variables over which to marganilize this probability's distribution
        :returns: A fraction in which the denominator is represents the sum over the given ranges

        >>> from y0.dsl import P, A, B
        >>> P(A, B).marginalize(A) == P(A, B) / Sum[A](P(A, B))
        """
        return self / Sum(expression=self, ranges=_prepare_ranges(ranges))

    def _iter_variables(self) -> Iterable[Variable]:
        """Get the set of variables used in the distribution in this probability."""
        yield from self.distribution._iter_variables()


def P(  # noqa:N802
    distribution: Union[Variable, Tuple[Variable, ...], Distribution],
    *args: Variable,
) -> Probability:
    """Create a probability expression over the given variable(s) or distribution.

    :param distribution: If given a :class:`Distribution`, creates a probability expression
        directly over the distribution. If given variable or list of variables, conveniently
        creates a :class:`Distribtion` with the variable(s) as children.
    :param args: If the first argument (``distribution``) was given as a single variable, the
        ``args`` variadic argument can be used to specify a list of additional variables.
    :returns: A probability object
    :raises ValueError: If varidic args are used incorrectly (i.e., in combination with a
        list of variables or :class:`Distribution`.

    Creation with a conditional distribution:

    >>> from y0.dsl import P, A, B
    >>> P(A | B)

    Creation with a joint distribution:

    >>> from y0.dsl import P, A, B
    >>> P(A & B)

    Creation with a mixed joint/conditional distribution:

    >>> from y0.dsl import P, A, B, C
    >>> P(A & B | C)

    Creation with a single :class:`Variable`:

    >>> from y0.dsl import P, A
    >>> P(A)

    Creation with a list of :class:`Variable`:

    >>> from y0.dsl import P, A, B
    >>> P((A, B))

    Creation with a list of :class:`Variable`: using variadic arguments:

    >>> from y0.dsl import P, A, B
    >>> P(A, B)
    """
    if isinstance(distribution, Variable):
        if not args:
            distribution = (distribution,)
        elif not all(isinstance(p, Variable) for p in args):
            raise ValueError
        else:
            distribution = (distribution, *args)
    if isinstance(distribution, tuple):
        distribution = Distribution(children=distribution)
    return Probability(distribution=distribution)


@dataclass(frozen=True)
class Product(Expression):
    """Represent the product of several probability expressions."""

    expressions: Tuple[Expression, ...]

    def to_text(self):
        """Output this product in the internal string format."""
        return " ".join(expression.to_text() for expression in self.expressions)

    def to_latex(self):
        """Output this product in the LaTeX string format."""
        return " ".join(expression.to_latex() for expression in self.expressions)

    def __mul__(self, other: Expression):
        if isinstance(other, Product):
            return Product((*self.expressions, *other.expressions))
        elif isinstance(other, Fraction):
            return Fraction(self * other.numerator, other.denominator)
        else:
            return Product((*self.expressions, other))

    def __truediv__(self, expression: Expression) -> Fraction:
        return Fraction(self, expression)

    def _iter_variables(self) -> Iterable[Variable]:
        """Get the union of the variables used in each expresison in this product."""
        for expression in self.expressions:
            yield from expression._iter_variables()


def _list_to_text(elements: Iterable[_Mathable]) -> str:
    return ",".join(element.to_text() for element in elements)


def _list_to_latex(elements: Iterable[_Mathable]) -> str:
    return ",".join(element.to_latex() for element in elements)


@dataclass(frozen=True)
class Sum(Expression):
    """Represent the sum over an expression over an optional set of variables."""

    #: The expression over which the sum is done
    expression: Expression
    #: The variables over which the sum is done. Defaults to an empty list, meaning no variables.
    ranges: Tuple[Variable, ...] = field(default_factory=tuple)

    def to_text(self) -> str:
        """Output this sum in the internal string format."""
        ranges = _list_to_text(self.ranges)
        return f"[ sum_{{{ranges}}} {self.expression.to_text()} ]"

    def to_latex(self) -> str:
        """Output this sum in the LaTeX string format."""
        ranges = _list_to_latex(self.ranges)
        return rf"\sum_{{{ranges}}} {self.expression.to_latex()}"

    def __mul__(self, expression: Expression):
        if isinstance(expression, Product):
            return Product((self, *expression.expressions))
        else:
            return Product((self, expression))

    def __truediv__(self, expression: Expression) -> Fraction:
        return Fraction(self, expression)

    def _iter_variables(self) -> Iterable[Variable]:
        """Get the union of the variables used in the range of this sum and variables in its summand."""
        yield from self.expression._iter_variables()
        for variable in self.ranges:
            yield from variable._iter_variables()

    @classmethod
    def __class_getitem__(
        cls, ranges: Union[Variable, Tuple[Variable, ...]]
    ) -> Callable[[Expression], Sum]:
        """Create a partial sum object over the given ranges.

        :param ranges: The variables over which the partial sum will be done
        :returns: A partial :class:`Sum` that can be called solely on an expression

        Example single variable sum:

        >>> from y0.dsl import Sum, P, A, B
        >>> Sum[B](P(A | B) * P(B))

        Example multiple variable sum:

        >>> from y0.dsl import Sum, P, A, B, C
        >>> Sum[B, C](P(A | B) * P(B))
        """
        return functools.partial(Sum, ranges=_prepare_ranges(ranges))


def _prepare_ranges(ranges: XSeq[Variable]) -> Tuple[Variable, ...]:
    if isinstance(
        ranges, Variable
    ):  # a single element is not given as a tuple, such as in Sum[T]
        return (ranges,)
    return tuple(ranges)


@dataclass(frozen=True)
class Fraction(Expression):
    """Represents a fraction of two expressions."""

    #: The expression in the numerator of the fraction
    numerator: Expression
    #: The expression in the denominator of the fraction
    denominator: Expression

    def to_text(self) -> str:
        """Output this fraction in the internal string format."""
        return f"frac_{{{self.numerator.to_text()}}}{{{self.denominator.to_text()}}}"

    def to_latex(self) -> str:
        """Output this fraction in the LaTeX string format."""
        return rf"\frac{{{self.numerator.to_latex()}}}{{{self.denominator.to_latex()}}}"

    def __mul__(self, expression: Expression) -> Fraction:
        if isinstance(expression, Fraction):
            return Fraction(
                self.numerator * expression.numerator,
                self.denominator * expression.denominator,
            )
        else:
            return Fraction(self.numerator * expression, self.denominator)

    def __truediv__(self, expression: Expression) -> Fraction:
        if isinstance(expression, Fraction):
            return Fraction(
                self.numerator * expression.denominator,
                self.denominator * expression.numerator,
            )
        elif isinstance(expression, One):
            return self
        else:
            return Fraction(self.numerator, self.denominator * expression)

    def _iter_variables(self) -> Iterable[Variable]:
        """Get the set of variables used in the numerator and denominator of this fraction."""
        yield from self.numerator._iter_variables()
        yield from self.denominator._iter_variables()

    def simplify(self) -> Expression:
        """Simplify this fraction."""
        if isinstance(self.denominator, One):
            return self.numerator
        if isinstance(self.numerator, One):
            return self
        if self.numerator == self.denominator:
            return One()
        if isinstance(self.numerator, Product) and isinstance(
            self.denominator, Product
        ):
            return self._simplify_parts(
                self.numerator.expressions, self.denominator.expressions
            )
        elif isinstance(self.numerator, Product):
            return self._simplify_parts(self.numerator.expressions, [self.denominator])
        elif isinstance(self.denominator, Product):
            return self._simplify_parts([self.numerator], self.denominator.expressions)
        return self

    @classmethod
    def _simplify_parts(
        cls, numerator: Sequence[Expression], denominator: Sequence[Expression]
    ) -> Expression:
        """Calculate the minimum fraction.

        :param numerator: A sequence of expressions that are multiplied in the product in the numerator
        :param denominator: A sequence of expressions that are multiplied in the product in the denominator
        :returns: A simplified fraction.
        """
        new_numerator, new_denominator = cls._simplify_parts_helper(
            numerator, denominator
        )
        if new_numerator and new_denominator:
            return Fraction(
                _expression_or_product(new_numerator),
                _expression_or_product(new_denominator),
            )
        elif new_numerator:
            return _expression_or_product(new_numerator)
        elif new_denominator:
            return One() / _expression_or_product(new_denominator)
        else:
            return One()

    @staticmethod
    def _simplify_parts_helper(
        numerator: Sequence[Expression],
        denominator: Sequence[Expression],
    ) -> Tuple[Tuple[Expression, ...], Tuple[Expression, ...]]:
        numerator_cancelled = set()
        denominator_cancelled = set()
        for i, n_expr in enumerate(numerator):
            for j, d_expr in enumerate(denominator):
                if j in denominator_cancelled:
                    continue
                if n_expr == d_expr:
                    numerator_cancelled.add(i)
                    denominator_cancelled.add(j)
                    break
        return (
            tuple(
                expr for i, expr in enumerate(numerator) if i not in numerator_cancelled
            ),
            tuple(
                expr
                for i, expr in enumerate(denominator)
                if i not in denominator_cancelled
            ),
        )


def _expression_or_product(e: Sequence[Expression]) -> Expression:
    if not e:
        raise ValueError
    if 1 == len(e):
        return e[0]
    return Product(tuple(e))


class One(Expression):
    """The multiplicative identity (1)."""

    def to_text(self) -> str:
        """Output this identity variable in the internal string format."""
        return "1"

    def to_latex(self) -> str:
        """Output this identity instance in the LaTeX string format."""
        return "1"

    def __rmul__(self, expression: Expression) -> Expression:
        return expression

    def __mul__(self, expression: Expression) -> Expression:
        return expression

    def __truediv__(self, other: Expression) -> Fraction:
        return Fraction(self, other)

    def __eq__(self, other):
        return isinstance(other, One)  # all ones are equal

    def _iter_variables(self) -> Iterable[Variable]:
        """Get the set of variables used in this expression."""
        return iter([])


@dataclass(frozen=True)
class QFactor(Expression):
    """A function from the variables in the domain to a probability function over variables in the codomain."""

    domain: Tuple[Variable, ...]
    codomain: Tuple[Variable, ...]

    @classmethod
    def __class_getitem__(cls, codomain: Union[Variable, Tuple[Variable, ...]]):
        """Create a partial Q Factor object over the given codomain.

        :param codomain: The variables over which the partial Q Factor will be done
        :returns: A partial :class:`QFactor` that can be called solely on an expression

        Example single variable codomain Q expression:

        >>> from y0.dsl import Sum, Q, A, B, C
        >>> Q[C](A, B)

        Example multiple variable codomain Q expression:

        >>> from y0.dsl import Sum, Q, A, B, C, D
        >>> Q[C, D](A, B)
        """

        def _helper(*domain: Variable):
            return QFactor(domain=domain, codomain=_prepare_ranges(codomain))

        return _helper

    def to_text(self) -> str:
        """Output this fraction in the internal string format."""
        codomain = _list_to_latex(self.codomain)
        domain = _list_to_text(self.domain)
        return f"Q[{codomain}]({domain})"

    def to_latex(self) -> str:
        """Output this fraction in the LaTeX string format."""
        codomain = _list_to_latex(self.codomain)
        domain = _list_to_text(self.domain)
        return rf"Q_{{{codomain}}}({{{domain}}})"

    def __mul__(self, other: Expression):
        if isinstance(other, Product):
            return Product((self, *other.expressions))
        elif isinstance(other, Fraction):
            return Fraction(self * other.numerator, other.denominator)
        else:
            return Product((self, other))

    def __truediv__(self, expression: Expression) -> Fraction:
        if isinstance(expression, Fraction):
            return Fraction(self * expression.denominator, expression.numerator)
        else:
            return Fraction(self, expression)

    def _iter_variables(self) -> Iterable[Variable]:
        yield from self.codomain
        yield from self.domain


Q = QFactor

A, B, C, D, R, S, T, W, X, Y, Z = map(Variable, "ABCDRSTWXYZ")  # type: ignore
V1, V2, V3, V4, V5, V6 = [Variable(f"V{i}") for i in range(1, 7)]
Z1, Z2, Z3, Z4, Z5, Z6 = [Variable(f"Z{i}") for i in range(1, 7)]


def _upgrade_ordering(variables: Iterable[Union[str, Variable]]) -> Sequence[Variable]:
    return tuple(Variable.norm(variable) for variable in variables)


OrderingHint = Optional[Sequence[Union[str, Variable]]]


def ensure_ordering(
    expression: Expression,
    *,
    ordering: OrderingHint = None,
) -> Sequence[Variable]:
    """Get a canonical ordering of the variables in the expression, or pass one through.

    The canonical ordering of the variables in a given expression is based on the alphabetical
    sort order of the variables based on their names.

    :param expression: The expression to get a canonical ordering from.
    :param ordering: A given ordering to pass through if not none, otherwise calculate it.
    :returns: The ordering
    """
    if ordering is not None:
        return _upgrade_ordering(ordering)
    # use alphabetical ordering
    return sorted(expression.get_variables(), key=attrgetter("name"))



# -*- coding: utf-8 -*-

"""Test the probability parser."""

import unittest

from pyparsing import ParserElement

from y0.dsl import A, B, C, P, Q, Sum, X, Y, Z
from y0.parser.craig.grammar import fraction_pe, grammar, sum_pe
from y0.parser.craig.utils import probability_pe, qfactor_pe, variable_pe


class TestGrammar(unittest.TestCase):
    """Tests for parsing probability expressions."""

    def assert_many(
        self, expressions, parser_element: ParserElement, direct: bool = True
    ):
        """Help testing many."""
        for expression in expressions:
            with self.subTest(
                expr=expression.to_text(), parser_element=parser_element.name
            ):
                self.assert_parse_equal(expression, parser_element, direct=direct)

    def assert_parse_equal(
        self, expression, parser_element: ParserElement, direct: bool = True
    ):
        """Help test parsing works round trip.

        :param expression: The DSL object to check
        :param parser_element: The relevant parser element for specific checking
        :param direct: If true, uses object equals checks. If false, uses stringification then string equals checks.
            Switch to false when things aren't working during development.
        """
        text = expression.to_text()
        parse_result = parser_element.parseString(text)
        result_expression = parse_result.asList()[0]
        self.assertIsInstance(result_expression, expression.__class__)
        if direct:
            self.assertEqual(
                expression,
                result_expression,
                msg=f"Mismatch\nExpected: {repr(expression)}\nActual:   {repr(result_expression)}",
            )
        else:
            self.assertEqual(text, result_expression.to_text())

    def test_variable(self):
        """Tests for variables, intervention variables, and counterfactual variables."""
        self.assert_many(
            [
                A,
                A @ B,
                A @ ~B,
                A @ B @ C,
                A @ ~B @ C,
                A @ ~B @ ~C,
                A @ B @ ~C,
            ],
            variable_pe,
        )

    def test_probability(self):
        """Tests for probabilities."""
        for g in (probability_pe, grammar):
            self.assert_many(
                [
                    P(A),
                    P(A, B),
                    P(A | B),
                    P(A @ X | B),
                    P(A @ ~X | B),
                    P(A | B @ Y),
                    P(A | B @ ~Y),
                    P(A @ X | B @ ~Y),
                    P(A @ X | B @ ~Y | C @ Z),
                    P(A | [B, C]),
                    P(A, B, C),
                ],
                g,
            )

    def test_sum(self):
        """Tests for sums."""
        for g in (sum_pe, grammar):
            self.assert_many(
                [
                    Sum(P(A)),
                    Sum(P(A, B)),
                    Sum(P(A | B)),
                    Sum(P(A | B) * P(B)),
                    Sum[B](P(A | B) * P(B)),
                    Sum[B, C](P(A | B) * P(B)),
                ],
                g,
            )

    def test_fraction(self):
        """Tests for fractions."""
        for g in (fraction_pe, grammar):
            self.assert_many(
                [
                    Sum(P(A)) / P(B),
                    Sum(P(A, B)) / P(A),
                    Sum[B](P(A | B) * P(B)) / Sum(P(A | B) * P(B)),
                ],
                g,
            )

    def test_q(self):
        """Tests for Q factors."""
        for g in (qfactor_pe, grammar):
            self.assert_many(
                [
                    Q[A](X),
                    Q[A, B](X),
                    Q[A](X, Y),
                    Q[A, B](X, Y),
                ],
                g,
            )



# -*- coding: utf-8 -*-

"""Tests for canceling fractions."""

import unittest

from y0.dsl import A, B, One, P, Sum

one = One()


class TestCancel(unittest.TestCase):
    """Test cancelling fractions."""

    def test_simple_identity(self):
        """Test cancelling when the numerator and denominator are the same."""
        for label, frac in [
            ("one", one / one),
            ("prob", P(A) / P(A)),
            ("sum", Sum(P(A)) / Sum(P(A))),
            ("product", (P(A) * P(B)) / (P(A) * P(B))),
        ]:
            with self.subTest(type=label):
                self.assertEqual(one, frac.simplify(), msg=f"\n\nActual:{frac}")

    def test_fraction_simplify(self):
        """Test cancelling on products."""
        for label, expected, frac in [
            ("leave num.", P(B), (P(A) * P(B)) / P(A)),
            ("leave den.", one / P(B), P(A) / (P(A) * P(B))),
            ("unordered", one, (P(A) * P(B)) / (P(B) * P(A))),
        ]:
            with self.subTest(type=label):
                self.assertEqual(expected, frac.simplify(), msg=f"\n\nActual:{frac}")



# -*- coding: utf-8 -*-

"""A parser for causaleffect probability expressions based on :mod:`pyparsing`."""

from pyparsing import (
    Group,
    Optional,
    ParseResults,
    Suppress,
    Word,
    alphanums,
    alphas,
    delimitedList,
    nums,
)

from ..craig.utils import _make_probability, _make_q, _unpack
from ...dsl import Variable


def _make_variable(_s, _l, tokens: ParseResults) -> Variable:
    name = tokens["name"]
    if "subscript" in tokens:
        name += "_" + tokens["subscript"]
    return Variable(name=name)


subscript = Suppress("_{") + Word(nums)("subscript") + Suppress("}")
variable_pe = Word(alphas, alphanums)("name") + Optional(subscript)
variable_pe.setParseAction(_make_variable)
variable_pe.setName("variable")

variables_pe = delimitedList(Group(variable_pe).setParseAction(_unpack))
_children_pe = Group(variables_pe).setResultsName("children")
_parents_pe = Group(Optional(Suppress("|") + variables_pe)).setResultsName("parents")
probability_pe = Suppress("P(") + _children_pe + _parents_pe + Suppress(")")
probability_pe.setParseAction(_make_probability)
probability_pe.setName("probability")

qfactor_pe = (
    Suppress("Q[\\{")
    + Group(variables_pe).setResultsName("codomain")
    + Suppress("\\}](")
    + Group(variables_pe).setResultsName("domain")
    + Suppress(")")
)
qfactor_pe.setParseAction(_make_q)
qfactor_pe.setName("qfactor")



# -*- coding: utf-8 -*-

"""Test getting conditional independencies (and related)."""

import unittest
from typing import Iterable, Set, Union

from ananke.graphs import ADMG, SG

from y0.algorithm.conditional_independencies import (
    are_d_separated,
    get_conditional_independencies,
    get_moral_links,
)
from y0.examples import Example, d_separation_example, examples
from y0.graph import NxMixedGraph
from y0.struct import DSeparationJudgement


class TestDSeparation(unittest.TestCase):
    """Test the d-separation utility."""

    def test_mit_example(self):
        """Test checking D-separation on the MIT example."""
        graph = d_separation_example.graph.to_admg()

        self.assertFalse(are_d_separated(graph, "AA", "B", conditions=["D", "F"]))
        self.assertTrue(are_d_separated(graph, "AA", "B"))
        self.assertTrue(are_d_separated(graph, "D", "E", conditions=["C"]))
        self.assertFalse(are_d_separated(graph, "AA", "B", conditions=["C"]))
        self.assertFalse(are_d_separated(graph, "D", "E"))
        self.assertFalse(are_d_separated(graph, "D", "E", conditions=["AA", "B"]))
        self.assertFalse(are_d_separated(graph, "G", "G", conditions=["C"]))

    def test_examples(self):
        """Check that example conditional independencies are d-separations and that conditions (if present) are required.

        This test is using convenient examples to ensure that the d-separation algorithm
        isn't just always returning true or false.
        """
        testable = (
            example
            for example in examples
            if example.conditional_independencies is not None
        )

        for example in testable:
            with self.subTest(name=example.name):
                graph = example.graph.to_admg()
                for ci in example.conditional_independencies:
                    self.assertTrue(
                        are_d_separated(
                            graph, ci.left, ci.right, conditions=ci.conditions
                        ),
                        msg="Expected d-separation not found",
                    )
                    if ci.conditions:
                        self.assertFalse(
                            are_d_separated(graph, ci.left, ci.right),
                            msg="Unexpected d-separation",
                        )

    def test_moral_links(self):
        """Test adding 'moral links' (part of the d-separation algorithm).

        This test covers several cases around moral links to ensure that they are added when needed.
        """
        graph = ADMG(
            vertices=("a", "b", "c"),
            di_edges=[("a", "b"), ("b", "c")],
        )
        links = set(tuple(sorted(e)) for e in get_moral_links(graph))
        self.assertEqual(set(), links, msg="Unexpected moral links added.")

        graph = ADMG(
            vertices=("a", "b", "c"),
            di_edges=[("a", "c"), ("b", "c")],
        )
        links = set(tuple(sorted(e)) for e in get_moral_links(graph))
        self.assertEqual(
            {("a", "b")}, links, msg="Moral links not as expected in single-link case."
        )

        graph = ADMG(
            vertices=("a", "b", "aa", "bb", "c"),
            di_edges=[("a", "c"), ("b", "c"), ("aa", "c"), ("bb", "c")],
        )
        links = set(tuple(sorted(e)) for e in get_moral_links(graph))
        self.assertEqual(
            {
                ("a", "b"),
                ("a", "aa"),
                ("a", "bb"),
                ("aa", "b"),
                ("aa", "bb"),
                ("b", "bb"),
            },
            links,
            msg="Moral links not as expected in multi-link case.",
        )

        graph = ADMG(
            vertices=("a", "b", "c", "d", "e"),
            di_edges=[("a", "c"), ("b", "c"), ("c", "e"), ("d", "e")],
        )
        links = set(tuple(sorted(e)) for e in get_moral_links(graph))
        self.assertEqual(
            {("a", "b"), ("c", "d")},
            links,
            msg="Moral links not as expected in multi-site case.",
        )


class TestGetConditionalIndependencies(unittest.TestCase):
    """Test getting conditional independencies."""

    def assert_example_has_judgements(self, example: Example) -> None:
        """Assert that the example is consistent w.r.t. D-separations."""
        self.assertIsNotNone(example.conditional_independencies)
        self.assert_has_judgements(
            graph=example.graph,
            judgements=example.conditional_independencies,
        )

    def assert_has_judgements(
        self, graph: Union[NxMixedGraph, SG], judgements: Iterable[DSeparationJudgement]
    ) -> None:
        """Assert that the graph has the correct conditional independencies.

        :param graph: the graph to test
        :param judgements: the set of expected conditional independencies
        """
        asserted_judgements = set(judgements)
        observed_judgements = get_conditional_independencies(graph)

        self.assertIsNotNone(asserted_judgements, "Expected independencies is empty.")
        self.assertIsNotNone(observed_judgements, "Observed independencies is empty.")
        self.assertTrue(
            all(judgement.is_canonical for judgement in observed_judgements),
            msg="one or more of the returned DSeparationJudgement instances are not canonical",
        )

        self.assert_valid_judgements(graph, asserted_judgements)
        self.assert_valid_judgements(graph, observed_judgements)

        expected_pairs = {
            (judgement.left, judgement.right) for judgement in asserted_judgements
        }
        observed_pairs = {
            (judgement.left, judgement.right) for judgement in observed_judgements
        }
        self.assertEqual(
            expected_pairs,
            observed_pairs,
            "Judgements do not find same separable pairs",
        )

        def _get_match(ref, options):
            """Find a judgement that has the same left/right pair as the reference judgement."""
            for alt in options:
                if ref.left == alt.left and ref.right == alt.right:
                    return alt
            return None

        for judgement in asserted_judgements:
            with self.subTest(name=judgement):
                matching = _get_match(judgement, observed_judgements)
                self.assertIsNotNone(matching, "No matching judgement found.")
                self.assertGreaterEqual(
                    len(judgement.conditions),
                    len(matching.conditions),
                    msg="Observed conditional independence more complicated than reference.",
                )

    def assert_valid_judgements(
        self, graph: Union[NxMixedGraph, SG], judgements: Set[DSeparationJudgement]
    ) -> None:
        """Check that a set of judgments are valid with respect to a graph."""
        if isinstance(graph, NxMixedGraph):
            graph = graph.to_admg()
        for judgement in judgements:
            self.assertTrue(
                are_d_separated(
                    graph,
                    judgement.left,
                    judgement.right,
                    conditions=judgement.conditions,
                ),
                msg=f"Conditional independency is not a d-separation in {graph}",
            )

        pairs = [(judgement.left, judgement.right) for judgement in judgements]
        self.assertEqual(
            len(pairs), len(set(pairs)), "Duplicate left/right pair observed"
        )

    def test_examples(self):
        """Test getting the conditional independencies from the example graphs."""
        testable = (
            example
            for example in examples
            if example.conditional_independencies is not None
        )

        for example in testable:
            with self.subTest(name=example.name):
                self.maxDiff = None
                self.assert_example_has_judgements(example)



# -*- coding: utf-8 -*-

"""Tests for chain mutations."""

import unittest

from y0.dsl import A, B, P, Sum, W, X, Y, Z
from y0.mutate import bayes_expand, chain_expand, fraction_expand


class TestChain(unittest.TestCase):
    """Tests for chain mutations."""

    def test_chain_expand_automatic_reordering(self):
        """Test expanding a joint probability to a product of conditional probabilities."""
        self.assertEqual(P(X | Y) * P(Y), chain_expand(P(X, Y), reorder=True))
        self.assertEqual(
            P(X | (Y, Z)) * P(Y | Z) * P(Z), chain_expand(P(X, Y, Z), reorder=True)
        )
        self.assertEqual(
            P(W | (X, Y, Z)) * P(X | (Y, Z)) * P(Y | Z) * P(Z),
            chain_expand(P(W, X, Y, Z), reorder=True),
        )

        # Test that conditions come along for the ride
        self.assertEqual(
            P(X | (Y, A)) * P(Y | A), chain_expand(P(X & Y | A), reorder=True)
        )
        self.assertEqual(
            P(X | (Y, A, B)) * P(Y | (A, B)),
            chain_expand(P(X & Y | (A, B)), reorder=True),
        )

    def test_chain_expand_no_reordering(self):
        """Test expanding a joint probability to a product of conditional probabilities."""
        self.assertEqual(P(X | Y) * P(Y), chain_expand(P(X, Y), reorder=False))
        self.assertEqual(
            P(X | (Y, Z)) * P(Y | Z) * P(Z), chain_expand(P(X, Y, Z), reorder=False)
        )
        self.assertEqual(
            P(W | (X, Y, Z)) * P(X | (Y, Z)) * P(Y | Z) * P(Z),
            chain_expand(P(W, X, Y, Z), reorder=False),
        )

    def test_fraction_expand(self):
        """Test expanding a conditional probability with Bayes' Theorem."""
        self.assertEqual(P(A, B) / P(B), fraction_expand(P(A | B)))
        self.assertEqual(P(W, X, Y, Z) / P(X, Y, Z), fraction_expand(P(W | (X, Y, Z))))

    def test_bayes_expand(self):
        """Test expanding a conditional using extended Bayes' Theorem."""
        self.assertEqual(P(A, X) / Sum[A](P(A, X)), bayes_expand(P(A | X)))
        self.assertEqual(P(A, X, Y) / Sum[A](P(A, X, Y)), bayes_expand(P(A | (X, Y))))
        self.assertEqual(
            P(A, B, X, Y) / Sum[A, B](P(A, B, X, Y)), bayes_expand(P(A & B | (X, Y)))
        )



# -*- coding: utf-8 -*-

"""Test graph construction and conversion."""

import unittest
from textwrap import dedent
from typing import Set, Tuple

from y0.examples import verma_1
from y0.graph import DEFAULT_TAG, DEFULT_PREFIX, NxMixedGraph
from y0.resources import VIRAL_PATHOGENESIS_PATH


class TestGraph(unittest.TestCase):
    """Test graph construction and conversion."""

    def test_causaleffect_str_verma_1(self):
        """Test generating R code for the figure 1A graph for causaleffect."""
        expected = dedent(
            """
        g <- graph.formula(V1 -+ V2, V2 -+ V3, V3 -+ V4, V2 -+ V4, V4 -+ V2, simplify = FALSE)
        g <- set.edge.attribute(graph = g, name = "description", index = c(4, 5), value = "U")
        """
        ).strip()
        self.assertEqual(expected, verma_1.to_causaleffect_str())

    def assert_labeled_convertable(
        self, graph: NxMixedGraph, labeled_edges: Set[Tuple[str, str]]
    ) -> None:
        """Test that the graph can be converted to a DAG, then back to an ADMG."""
        prefix = DEFULT_PREFIX
        tag = DEFAULT_TAG

        labeled_dag = graph.to_latent_variable_dag(prefix=prefix, tag=tag)
        for node in labeled_dag:
            self.assertIn(tag, labeled_dag.nodes[node], msg=f"Node: {node}")
            self.assertEqual(node.startswith(prefix), labeled_dag.nodes[node][tag])

        self.assertEqual(labeled_edges, set(labeled_dag.edges()))

        reconstituted = NxMixedGraph.from_latent_variable_dag(labeled_dag, tag=tag)
        self.assertEqual(
            set(graph.directed.nodes()), set(reconstituted.directed.nodes())
        )
        self.assertEqual(
            set(graph.undirected.nodes()), set(reconstituted.undirected.nodes())
        )
        self.assertEqual(
            set(graph.directed.edges()), set(reconstituted.directed.edges())
        )
        self.assertEqual(
            set(graph.undirected.edges()), set(reconstituted.undirected.edges())
        )

    def test_convertable(self):
        """Test graphs are convertable."""
        for graph, labeled_edges in [
            (
                verma_1,
                {
                    ("V1", "V2"),
                    ("V2", "V3"),
                    ("V3", "V4"),
                    (f"{DEFULT_PREFIX}0", "V2"),
                    (f"{DEFULT_PREFIX}0", "V4"),
                },
            ),
        ]:
            with self.subTest():
                self.assert_labeled_convertable(graph, labeled_edges)

    def test_from_causalfusion(self):
        """Test importing a CausalFusion graph."""
        graph = NxMixedGraph.from_causalfusion_path(VIRAL_PATHOGENESIS_PATH)
        self.assertIsInstance(graph, NxMixedGraph)



# -*- coding: utf-8 -*-

"""Implementation of the canonicalization algorithm."""

from typing import Mapping, Optional, Sequence, Tuple, Union

from ..dsl import (
    CounterfactualVariable,
    Distribution,
    Expression,
    Fraction,
    Intervention,
    Probability,
    Product,
    Sum,
    Variable,
    ensure_ordering,
)

__all__ = [
    "canonicalize",
]


def canonicalize(
    expression: Expression, ordering: Optional[Sequence[Union[str, Variable]]] = None
) -> Expression:
    """Canonicalize an expression that meets the markov condition with respect to the given ordering.

    :param expression: An expression to canonicalize
    :param ordering: A toplogical ordering. If none is given, it is assigned by sort order of the variable names.
    :return: A canonical expression
    """
    canonicalizer = Canonicalizer(ensure_ordering(expression, ordering=ordering))
    return canonicalizer.canonicalize(expression)


def _sort_probability_key(probability: Probability) -> str:
    return probability.distribution.children[0].name


class Canonicalizer:
    """A data structure to support application of the canonicalize algorithm."""

    ordering: Sequence[Variable]
    ordering_level: Mapping[str, int]

    def __init__(self, ordering: Sequence[Variable]) -> None:
        """Initialize the canonicalizer.

        :param ordering: A topological ordering over the variables appearing in the expression.

        :raises ValueError: if the ordering has duplicates
        """
        if len(set(ordering)) != len(ordering):
            raise ValueError(f"ordering has duplicates: {ordering}")

        self.ordering = ordering
        self.ordering_level = {
            variable.name: level for level, variable in enumerate(self.ordering)
        }

    def _canonicalize_probability(self, expression: Probability) -> Probability:
        return Probability(
            Distribution(
                children=self._sorted(expression.distribution.children),
                parents=self._sorted(expression.distribution.parents),
            )
        )

    def _sorted(self, variables: Tuple[Variable, ...]) -> Tuple[Variable, ...]:
        return tuple(
            sorted(
                (self._canonicalize_variable(variable) for variable in variables),
                key=self._sorted_key,
            )
        )

    def _canonicalize_variable(self, variable: Variable) -> Variable:
        if isinstance(variable, CounterfactualVariable):
            return CounterfactualVariable(
                name=variable.name,
                interventions=tuple(
                    sorted(variable.interventions, key=self._intervention_key)
                ),
            )
        else:
            return variable

    @staticmethod
    def _intervention_key(intervention: Intervention):
        return intervention.name, intervention.star

    def _sorted_key(self, variable: Variable) -> int:
        return self.ordering_level[variable.name]

    def canonicalize(self, expression: Expression) -> Expression:
        """Canonicalize an expression.

        :param expression: An uncanonicalized expression
        :return: A canonicalized expression
        :raises TypeError: if an object with an invalid type is passed
        """
        if isinstance(expression, Probability):  # atomic
            return self._canonicalize_probability(expression)
        elif isinstance(expression, Sum):
            return Sum(
                expression=self.canonicalize(expression.expression),
                ranges=expression.ranges,
            )
        elif isinstance(expression, Product):
            probabilities = []
            other = []
            for subexpr in expression.expressions:
                subexpr = self.canonicalize(subexpr)
                if isinstance(subexpr, Probability):
                    probabilities.append(subexpr)
                else:
                    other.append(subexpr)
            probabilities = sorted(probabilities, key=_sort_probability_key)

            # If other is empty, this is also atomic
            other = sorted(other, key=self._nonatomic_key)
            return Product((*probabilities, *other))
        elif isinstance(expression, Fraction):
            return Fraction(
                numerator=self.canonicalize(expression.numerator),
                denominator=self.canonicalize(expression.denominator),
            )
        else:
            raise TypeError

    def _nonatomic_key(self, expression: Expression):
        """Generate a sort key for a *canonical* expression.

        :param expression: A canonical expression
        :returns: A tuple in which the first element is the integer priority for the expression
            and the rest depends on the expression type.
        :raises TypeError: if an invalid expression type is given
        """
        if isinstance(expression, Probability):
            return 0, expression.distribution.children[0].name
        elif isinstance(expression, Sum):
            return 1, *self._nonatomic_key(expression.expression)
        elif isinstance(expression, Product):
            inner_keys = (
                self._nonatomic_key(sexpr) for sexpr in expression.expressions
            )
            return 2, *inner_keys
        elif isinstance(expression, Fraction):
            return (
                3,
                self._nonatomic_key(expression.numerator),
                self._nonatomic_key(expression.denominator),
            )
        else:
            raise TypeError



# -*- coding: utf-8 -*-

"""Predicates for expressions."""

from .dsl import Expression, Fraction, Probability, Product, Sum

__all__ = [
    "has_markov_postcondition",
]


def has_markov_postcondition(expression: Expression) -> bool:
    """Check that the expression is a sum/product of markov kernels.

    :param expression: Any expression
    :return: if the expression satisfies the sum/product of markov kernels condition
    :raises TypeError: if an object with an invalid type is passed
    """
    if isinstance(expression, Probability):
        return expression.distribution.is_markov_kernel()
    elif isinstance(expression, Product):
        return all(
            has_markov_postcondition(subexpr) for subexpr in expression.expressions
        )
    elif isinstance(expression, Sum):
        return has_markov_postcondition(expression.expression)
    elif isinstance(expression, Fraction):
        return has_markov_postcondition(
            expression.numerator
        ) and has_markov_postcondition(expression.denominator)
    else:
        raise TypeError



# -*- coding: utf-8 -*-

"""Causal graphs have implications that can be tested in the context of a specific dataset.

This module includes algorithms to perform those tests.
"""

from collections import abc
from typing import Iterable, Optional, Union

import pandas as pd
from ananke.graphs import SG
from tqdm import tqdm

from .conditional_independencies import get_conditional_independencies
from ..struct import DSeparationJudgement
from ..util.stat_utils import cressie_read


class Falsifications(abc.Sequence):
    """A list of variables pairs that failed the D-separation and covariance test.

    Has an extra 'evidence' property that is a dictionary.

    - Keys are the d-separated variable pairs
    - Values are the covariances measured between them.
    """

    def __init__(self, failures, evidence: pd.DataFrame):
        """Create Falsifications result.

        :param failures: Sequence of implications that did not pass
        :param evidence: Collection of all implications tested
        """
        self._failures = failures
        self.evidence = evidence

    def __getitem__(self, i):
        return self._failures[i]

    def __len__(self):
        return len(self._failures)

    def __repr__(self):
        return repr(self._failures) + "+evidence"


def falsifications(
    to_test: Union[SG, Iterable[DSeparationJudgement]],
    df: pd.DataFrame,
    significance_level: float = 0.05,
    max_given: Optional[int] = None,
    verbose: bool = False,
) -> Falsifications:
    """Test conditional independencies implied by a graph.

    :param to_test: Either a graph to generate d-separation from or a list of D-separations to check.
    :param df: Data to check for consistency with a causal implications
    :param significance_level: Significance for p-value test
    :param max_given: The maximum set size in the power set of the vertices minus the d-separable pairs
    :param verbose: If true, use tqdm for status updates.
    :return: Falsifications report
    """
    if isinstance(to_test, SG):
        to_test = get_conditional_independencies(
            to_test, max_conditions=max_given, verbose=verbose
        )

    variances = {
        (left, right, given): cressie_read(left, right, given, df, boolean=False)
        for _, left, right, given in tqdm(
            to_test, disable=not verbose, desc="Checking conditionals"
        )
    }

    rows = [
        (left, right, given, chi, p, dof)
        for (left, right, given), (chi, dof, p) in variances.items()
    ]

    evidence = (
        pd.DataFrame(rows, columns=["left", "right", "given", "chi^2", "p", "dof"])
        .sort_values("p")
        .assign(
            **{
                "Holm–Bonferroni level": significance_level
                / pd.Series(range(len(rows) + 1, 0, -1))
            }
        )
        .pipe(_assign_flags)
        .sort_values(["flagged", "dof"], ascending=False)
    )

    failures = evidence[evidence["flagged"]][["left", "right", "given"]].apply(
        tuple, axis="columns"
    )
    return Falsifications(failures, evidence)


def _assign_flags(df: pd.DataFrame) -> pd.DataFrame:
    return df.assign(flagged=(df["p"] < df["Holm–Bonferroni level"]))



# -*- coding: utf-8 -*-

"""Tests for simplifying latent variable DAGs."""

import unittest

import networkx as nx

from y0.algorithm.simplify_latent import (
    DEFAULT_SUFFIX,
    iter_latents,
    remove_redundant_latents,
    remove_widow_latents,
    transform_latents_with_parents,
)
from y0.graph import set_latent


class TestSimplify(unittest.TestCase):
    """Tests for the Robin Evans simplification algorithms."""

    def assert_latent_variable_dag_equal(self, expected, actual) -> None:
        """Check two latent variable DAGs are the same."""
        self.assertEqual(
            sorted(expected),
            sorted(actual),
            msg="Nodes are incorrect",
        )
        self.assertEqual(
            dict(expected.nodes.items()),
            dict(actual.nodes.items()),
            msg="Tags are incorrect",
        )
        self.assertEqual(
            sorted(expected.edges()),
            sorted(actual.edges()),
        )

    def test_remove_widows(self):
        """Test simplification 1 - removing widows."""
        graph = nx.DiGraph()
        graph.add_edges_from(
            [
                ("X1", "X2"),
                ("X2", "W"),
                ("U", "X2"),
                ("U", "X3"),
                ("U", "W"),
            ]
        )
        latents = {"U", "W"}
        set_latent(graph, latents)
        self.assertEqual(latents, set(iter_latents(graph)))

        # Apply the simplification
        _, removed = remove_widow_latents(graph)
        self.assertEqual({"W"}, removed)

        expected = nx.DiGraph()
        expected.add_edges_from(
            [
                ("X1", "X2"),
                ("U", "X2"),
                ("U", "X3"),
            ]
        )
        set_latent(expected, "U")

        self.assert_latent_variable_dag_equal(expected, graph)

    def test_transform_latents_with_parents(self):
        """Test simplification 2: latents with parents can be transformed."""
        graph = nx.DiGraph()
        graph.add_edges_from(
            [
                ("X1", "U"),
                ("X2", "U"),
                ("U", "Y1"),
                ("U", "Y2"),
                ("U", "Y3"),
            ]
        )
        set_latent(graph, "U")
        # Apply the simplification
        transform_latents_with_parents(graph)

        expected = nx.DiGraph()
        expected.add_edges_from(
            [
                ("X1", "Y1"),
                ("X1", "Y2"),
                ("X1", "Y3"),
                ("X2", "Y1"),
                ("X2", "Y2"),
                ("X2", "Y3"),
                (f"U{DEFAULT_SUFFIX}", "Y1"),
                (f"U{DEFAULT_SUFFIX}", "Y2"),
                (f"U{DEFAULT_SUFFIX}", "Y3"),
            ]
        )
        set_latent(expected, f"U{DEFAULT_SUFFIX}")

        self.assert_latent_variable_dag_equal(expected, graph)

    def test_remove_redundant_latents(self):
        """Test simplification 3 - remove redundant latents."""
        graph = nx.DiGraph()
        graph.add_edges_from(
            [
                ("U", "X1"),
                ("U", "X2"),
                ("U", "X3"),
                ("W", "X1"),
                ("W", "X2"),
            ]
        )
        set_latent(graph, ["U", "W"])
        # Apply the simplification
        remove_redundant_latents(graph)

        expected = nx.DiGraph()
        expected.add_edges_from(
            [
                ("U", "X1"),
                ("U", "X2"),
                ("U", "X3"),
            ]
        )
        set_latent(expected, ["U"])

        self.assert_latent_variable_dag_equal(expected, graph)



# -*- coding: utf-8 -*-

"""Tests for the identify algorithm."""

import unittest
from typing import Union

from y0.dsl import Distribution, P, Probability, X, Y, Z
from y0.graph import NxMixedGraph
from y0.identify import _get_to, is_identifiable


class TestUtils(unittest.TestCase):
    """Test utility functions for ID algorithms."""

    def test_to(self):
        """Test getting treatments and outcomes."""
        with self.assertRaises(ValueError):
            _get_to(P(X).distribution)
        with self.assertRaises(ValueError):
            _get_to(P(X, Y).distribution)
        with self.assertRaises(ValueError):
            _get_to(P(X @ Y, X @ Z).distribution)

        for expected_t, expected_o, expression in [
            (["X"], ["Y"], P(Y @ X)),
            (["X"], ["Y", "Z"], P(Y @ X, Z @ X)),
        ]:
            t, o = _get_to(expression.distribution)
            self.assertEqual(expected_t, t)
            self.assertEqual(expected_o, o)


class TestNotIdentifiable(unittest.TestCase):
    """Tests for lack of identifiability.

    These tests are based on the examples from the Figure 1 series on
    https://github.com/COVID-19-Causal-Reasoning/Y0/blob/master/ID_whittemore.ipynb.
    """

    def assert_not_identifiable(
        self, graph: NxMixedGraph, query: Union[Probability, Distribution]
    ) -> None:
        """Asset the graph is not identifiable under the given query."""
        self.assertFalse(is_identifiable(graph, query))

    def test_figure_1a(self):
        """Test Figure 1A."""
        graph_1a = NxMixedGraph()
        graph_1a.add_directed_edge("X", "Y")
        graph_1a.add_undirected_edge("X", "Y")
        self.assert_not_identifiable(graph_1a, P(Y @ ~X))

    def test_figure_1b(self):
        """Test Figure 1B."""
        graph_1b = NxMixedGraph()
        graph_1b.add_directed_edge("X", "Z")
        graph_1b.add_directed_edge("Z", "Y")
        graph_1b.add_undirected_edge("X", "Z")
        self.assert_not_identifiable(graph_1b, P(Y @ ~X))

    def test_figure_1c(self):
        """Test Figure 1c."""
        graph_1c = NxMixedGraph()
        graph_1c.add_directed_edge("X", "Z")
        graph_1c.add_directed_edge("Z", "Y")
        graph_1c.add_directed_edge("X", "Y")
        graph_1c.add_undirected_edge("X", "Z")
        self.assert_not_identifiable(graph_1c, P(Y @ ~X))

    def test_figure_1d(self):
        """Test Figure 1d."""
        graph_1d = NxMixedGraph()
        graph_1d.add_directed_edge("X", "Y")
        graph_1d.add_directed_edge("Z", "Y")
        graph_1d.add_undirected_edge("X", "Z")
        graph_1d.add_undirected_edge("Z", "Y")
        self.assert_not_identifiable(graph_1d, P(Y @ ~X))

    def test_figure_1e(self):
        """Test Figure 1e."""
        graph_1e = NxMixedGraph()
        graph_1e.add_directed_edge("Z", "X")
        graph_1e.add_directed_edge("X", "Y")
        graph_1e.add_undirected_edge("X", "Z")
        graph_1e.add_undirected_edge("Z", "Y")
        self.assert_not_identifiable(graph_1e, P(Y @ ~X))

    def test_figure_1f(self):
        """Test Figure 1f."""
        graph_1f = NxMixedGraph()
        graph_1f.add_directed_edge("X", "Z")
        graph_1f.add_directed_edge("Z", "Y")
        graph_1f.add_undirected_edge("X", "Y")
        graph_1f.add_undirected_edge("Z", "Y")
        self.assert_not_identifiable(graph_1f, P(Y @ ~X))

    def test_figure_1g(self):
        """Test Figure 1g."""
        graph_1g = NxMixedGraph()
        graph_1g.add_directed_edge("X", "Z1")
        graph_1g.add_directed_edge("Z1", "Y")
        graph_1g.add_directed_edge("Z2", "Y")
        graph_1g.add_undirected_edge("X", "Z2")
        graph_1g.add_undirected_edge("Z1", "Z2")
        self.assert_not_identifiable(graph_1g, P(Y @ ~X))

    def test_figure_1h(self):
        """Test Figure 1h."""
        graph_1h = NxMixedGraph()
        graph_1h.add_directed_edge("Z", "X")
        graph_1h.add_directed_edge("X", "W")
        graph_1h.add_directed_edge("W", "Y")
        graph_1h.add_undirected_edge("X", "Z")
        graph_1h.add_undirected_edge("X", "Y")
        graph_1h.add_undirected_edge("W", "Z")
        graph_1h.add_undirected_edge("Y", "Z")
        self.assert_not_identifiable(graph_1h, P(Y @ ~X))


class TestIdentifiable(unittest.TestCase):
    """Tests for lack of identifiability.

    These tests are based on the examples from the Figure 2 series on
    https://github.com/COVID-19-Causal-Reasoning/Y0/blob/master/ID_whittemore.ipynb.
    """

    def assert_identifiable(
        self, graph: NxMixedGraph, query: Union[Probability, Distribution]
    ) -> None:
        """Assert the graph is identifiable under the given query."""
        self.assertTrue(is_identifiable(graph, query))

    def test_figure_2a(self):
        """Test Figure 2a."""
        graph_2a = NxMixedGraph()
        graph_2a.add_directed_edge("X", "Y")
        self.assert_identifiable(graph_2a, P(Y @ ~X))

    def test_figure_2b(self):
        """Test Figure 2B."""
        graph_2b = NxMixedGraph()
        graph_2b.add_directed_edge("X", "Y")
        graph_2b.add_directed_edge("X", "Z")
        graph_2b.add_directed_edge("Z", "Y")
        graph_2b.add_undirected_edge("Y", "Z")
        self.assert_identifiable(graph_2b, P(Y @ ~X))

    def test_figure_2c(self):
        """Test Figure 2C."""
        graph_2c = NxMixedGraph()
        graph_2c.add_directed_edge("X", "Y")
        graph_2c.add_directed_edge("Z", "X")
        graph_2c.add_directed_edge("Z", "Y")
        graph_2c.add_undirected_edge("Y", "Z")
        self.assert_identifiable(graph_2c, P(Y @ ~X))

    def test_figure_2d(self):
        """Test Figure 2D."""
        graph_2d = NxMixedGraph()
        graph_2d.add_directed_edge("X", "Y")
        graph_2d.add_directed_edge("Z", "X")
        graph_2d.add_directed_edge("Z", "Y")
        graph_2d.add_undirected_edge("X", "Z")
        self.assert_identifiable(graph_2d, P(Y @ ~X))

    def test_figure_2e(self):
        """Test Figure 2E."""
        graph_2e = NxMixedGraph()
        graph_2e.add_directed_edge("X", "Z")
        graph_2e.add_directed_edge("Z", "Y")
        graph_2e.add_undirected_edge("X", "Y")
        self.assert_identifiable(graph_2e, P(Y @ ~X))

    def test_figure_2f(self):
        """Test Figure 2f."""
        graph_2f = NxMixedGraph()
        graph_2f.add_directed_edge("X", "Y")
        graph_2f.add_directed_edge("X", "Z1")
        graph_2f.add_directed_edge("Z1", "Y")
        graph_2f.add_directed_edge("Z1", "Z2")
        graph_2f.add_directed_edge("Z2", "Y")
        graph_2f.add_undirected_edge("X", "Z2")
        graph_2f.add_undirected_edge("Y", "Z1")
        self.assert_identifiable(graph_2f, P(Y @ ~X))

    def test_figure_2g(self):
        """Test Figure 2g."""
        graph_2g = NxMixedGraph()
        graph_2g.add_directed_edge("Z2", "Z1")
        graph_2g.add_directed_edge("Z2", "X")
        graph_2g.add_directed_edge("Z2", "Z3")
        graph_2g.add_directed_edge("X", "Z1")
        graph_2g.add_directed_edge("Z1", "Y")
        graph_2g.add_directed_edge("Z3", "Y")
        graph_2g.add_undirected_edge("Z2", "X")
        graph_2g.add_undirected_edge("Z2", "Y")
        graph_2g.add_undirected_edge("X", "Z3")
        graph_2g.add_undirected_edge("X", "Y")
        self.assert_identifiable(graph_2g, P(Y @ ~X))



# -*- coding: utf-8 -*-

"""Tests for the canonicalization algorithm."""

import itertools as itt
import unittest
from typing import Sequence

from y0.dsl import A, B, C, D, Expression, P, Sum, Variable, X, Y, Z
from y0.mutate import canonicalize


class TestCanonicalize(unittest.TestCase):
    """Tests for the canonicalization of a simplified algorithm."""

    def assert_canonicalize(
        self, expected: Expression, expression: Expression, ordering: Sequence[Variable]
    ) -> None:
        """Check that the expression is canonicalized properly given an ordering."""
        with self.subTest(
            expr=str(expression),
            ordering=", ".join(variable.name for variable in ordering),
        ):
            actual = canonicalize(expression, ordering)
            self.assertEqual(
                expected,
                actual,
                msg=f"\nExpected: {str(expression)}\nActual:   {str(actual)}",
            )

    def test_atomic(self):
        """Test canonicalization of atomic expressions."""
        for expected, expression, ordering in [
            (P(A), P(A), [A]),
            (P(A | B), P(A | B), [A, B]),
            (P(A | (B, C)), P(A | (B, C)), [A, B, C]),
            (P(A | (B, C)), P(A | (C, B)), [A, B, C]),
        ]:
            self.assert_canonicalize(expected, expression, ordering)

        expected = P(A | (B, C, D))
        for b, c, d in itt.permutations((B, C, D)):
            expression = P(A | (b, c, d))
            self.assert_canonicalize(expected, expression, [A, B, C, D])

    def test_atomic_interventions(self):
        """Test canonicalization of atomic expressions containing interventions."""
        for expected, expression, ordering in [
            (P(A @ X), P(A @ X), [A, X]),
            (P(A @ [X, Y]), P(A @ [X, Y]), [A, X, Y]),
            (P(A @ [X, Y]), P(A @ [Y, X]), [A, X, Y]),
        ]:
            self.assert_canonicalize(expected, expression, ordering)

    def test_derived_atomic(self):
        """Test canonicalizing."""
        # Sum
        expected = expression = Sum(P(A))
        self.assert_canonicalize(expected, expression, [A])

        # Simple product (only atomic)
        expected = P(A) * P(B) * P(C)
        for a, b, c in itt.permutations((P(A), P(B), P(C))):
            expression = a * b * c
            self.assert_canonicalize(expected, expression, [A, B, C])

        # Sum with simple product (only atomic)
        expected = Sum(P(A) * P(B) * P(C))
        for a, b, c in itt.permutations((P(A), P(B), P(C))):
            expression = Sum(a * b * c)
            self.assert_canonicalize(expected, expression, [A, B, C])

        # Fraction
        expected = expression = P(A) / P(B)
        self.assert_canonicalize(expected, expression, [A, B])

        # Fraction with simple products (only atomic)
        expected = (P(A) * P(B) * P(C)) / (P(X) * P(Y) * P(Z))
        for (a, b, c), (x, y, z) in itt.product(
            itt.permutations((P(A), P(B), P(C))),
            itt.permutations((P(X), P(Y), P(Z))),
        ):
            expression = (a * b * c) / (x * y * z)
            self.assert_canonicalize(expected, expression, [A, B, C, X, Y, Z])

    def test_mixed(self):
        """Test mixed expressions."""
        expected = expression = P(A) * Sum(P(B))
        self.assert_canonicalize(expected, expression, [A, B])

        expected = P(A) * Sum(P(B)) * Sum(P(C))
        for a, b, c in itt.permutations((P(A), Sum(P(B)), Sum(P(C)))):
            expression = a * b * c
            self.assert_canonicalize(expected, expression, [A, B, C])

        expected = P(D) * Sum(P(A) * P(B) * P(C))
        for a, b, c in itt.permutations((P(A), P(B), P(C))):
            sum_expr = Sum(a * b * c)
            for left, right in itt.permutations((P(D), sum_expr)):
                self.assert_canonicalize(expected, left * right, [A, B, C, D])

        expected = P(X) * Sum(P(A) * P(B)) * Sum(P(C) * P(D))
        for (a, b), (c, d) in itt.product(
            itt.permutations((P(A), P(B))),
            itt.permutations((P(C), P(D))),
        ):
            sexpr = Sum(a * b) * Sum(c * d)
            self.assert_canonicalize(expected, sexpr * P(X), [A, B, C, D, X])
            self.assert_canonicalize(expected, P(X) * sexpr, [A, B, C, D, X])

        expected = expression = Sum(P(A) / P(B))
        self.assert_canonicalize(expected, expression, [A, B])

        expected = expression = Sum(P(A) / Sum(P(B))) * Sum(P(A) / Sum(P(B) / P(C)))
        self.assert_canonicalize(expected, expression, [A, B, C])

    def test_non_markov(self):
        """Test non-markov distributions (e.g., with multiple children)."""
        for c1, c2 in itt.permutations([A, B]):
            # No conditions
            self.assert_canonicalize(P(A & B), P(c1 & c2), [A, B])
            # One condition, C
            self.assert_canonicalize(P(A & B | C), P(c1 & c2 | C), [A, B, C])
            # Two conditions, C and D
            for p1, p2 in itt.permutations([C, D]):
                expected = P(A & B | C | D)
                expression = P(c1 & c2 | (p1, p2))
                ordering = [A, B, C, D]
                self.assert_canonicalize(expected, expression, ordering)
                self.assert_canonicalize(Sum(expected), Sum(expression), ordering)

        for c1, c2, c3 in itt.permutations([A, B, C]):
            self.assert_canonicalize(P(A, B, C), P(c1, c2, c3), [A, B, C])
            for p1, p2, p3 in itt.permutations([X, Y, Z]):
                expected = P(A & B & C | (X, Y, Z))
                expression = P(c1 & c2 & c3 | (p1 & p2 & p3))
                ordering = [A, B, C, X, Y, Z]
                self.assert_canonicalize(expected, expression, ordering)
                self.assert_canonicalize(Sum(expected), Sum(expression), ordering)



# -*- coding: utf-8 -*-

"""Test the probability DSL."""

import itertools as itt
import unittest

from y0.dsl import (
    A,
    B,
    C,
    CounterfactualVariable,
    D,
    Distribution,
    Fraction,
    Intervention,
    One,
    P,
    Q,
    R,
    S,
    Sum,
    T,
    Variable,
    W,
    X,
    Y,
    Z,
)

V = Variable("V")


class TestDSL(unittest.TestCase):
    """Tests for the stringifying instances of the probability DSL."""

    def assert_text(self, s: str, expression):
        """Assert the expression when it is converted to a string."""
        self.assertIsInstance(s, str)
        self.assertIsInstance(hash(expression), int)  # can the expression be hashed?
        self.assertIsInstance(expression.to_text(), str)
        self.assertIsInstance(expression.to_latex(), str)
        self.assertIsInstance(expression._repr_latex_(), str)
        self.assertEqual(s, expression.to_text(), msg=f"Expression: {repr(expression)}")

    def test_variable(self):
        """Test the variable DSL object."""
        self.assert_text("A", Variable("A"))
        self.assert_text("A", A)  # shorthand for testing purposes

    def test_stop_the_madness(self):
        """Test that a variable can not be named "P"."""
        with self.assertRaises(ValueError):
            _ = Variable("P")

    def test_intervention(self):
        """Test the invervention DSL object."""
        self.assert_text("W*", Intervention("W", True))
        self.assert_text("W", Intervention("W", False))
        self.assert_text("W", Intervention("W"))  # False is the default
        self.assert_text("W", W)  # shorthand for testing purposes
        self.assert_text("W", -W)  # An intervention from variable W

        # inversions using the unary ~ operator
        self.assert_text("W", ~Intervention("W", True))
        self.assert_text("W*", ~Intervention("W", False))  # False is still the default
        self.assert_text("W*", ~Intervention("W"))
        self.assert_text("W*", ~W)

    def test_counterfactual_variable(self):
        """Test the Counterfactual Variable DSL object."""
        # Normal instantiation
        self.assert_text("Y_{W}", CounterfactualVariable("Y", (-W,)))
        self.assert_text("Y_{W*}", CounterfactualVariable("Y", (~W,)))

        # Instantiation with list-based operand to matmul @ operator
        self.assert_text("Y_{W}", Variable("Y") @ [W])
        self.assert_text("Y_{W}", Y @ [W])
        self.assert_text("Y_{W*}", Variable("Y") @ [~W])
        self.assert_text("Y_{W*}", Y @ [~W])

        # Instantiation with two variables
        self.assert_text(
            "Y_{X,W*}",
            CounterfactualVariable("Y", (Intervention("X"), ~Intervention("W"))),
        )

        # Instantiation with matmul @ operator and single operand
        self.assert_text("Y_{W}", Y @ Intervention("W"))
        self.assert_text("Y_{W*}", Y @ ~Intervention("W"))

        # Instantiation with matmul @ operator and list operand
        self.assert_text("Y_{X,W*}", Y @ [X, ~W])

        # Instantiation with matmul @ operator (chained)
        self.assert_text("Y_{X,W*}", Y @ X @ ~W)

    def test_counterfactual_errors(self):
        """Test that if two variables with the same name are given, an error is raised, regardless of star state."""
        for a, b in itt.product([True, False], [True, False]):
            with self.subTest(a=a, b=b), self.assertRaises(ValueError):
                Y @ Intervention("X", star=a) @ Intervention("X", star=b)

    def test_conditional_distribution(self):
        """Test the :class:`Distribution` DSL object."""
        # Normal instantiation
        self.assert_text("A|B", Distribution((A,), (B,)))

        # Instantiation with list-based operand to or | operator
        self.assert_text("A|B", Variable("A") | (B,))
        self.assert_text("A|B", A | (B,))

        # # Instantiation with two variables
        self.assert_text("A|B,C", A | [B, C])

        # Instantiation with or | operator and single operand
        self.assert_text("A|B", Variable("A") | B)
        self.assert_text("A|B", A | B)

        # Instantiation with or | operator (chained)
        self.assert_text("A|B,C", A | B | C)

        # Counterfactual uses work basically the same.
        #  Note: @ binds more tightly than |, but it's probably better to use parentheses
        self.assert_text("Y_{W}|B", (Y @ W) | B)
        self.assert_text("Y_{W}|B", Y @ W | B)
        self.assert_text("Y_{W}|B,C", Y @ W | B | C)
        self.assert_text("Y_{W,X*}|B,C", Y @ W @ ~X | B | C)
        self.assert_text(
            "Y_{W,X*}|B_{N*},C", Y @ W @ ~X | B @ Intervention("N", True) | C
        )

    def test_joint_distribution(self):
        """Test the JointProbability DSL object."""
        self.assert_text("A,B", Distribution((A, B)))
        self.assert_text("A,B", A & B)
        self.assert_text("A,B,C", Distribution((A, B, C)))
        self.assert_text("A,B,C", A & B & C)

    def test_probability(self):
        """Test generation of probabilities."""
        # Make sure there are children
        with self.assertRaises(ValueError):
            Distribution(tuple())

        # Test markov kernels (AKA has only one child variable)
        self.assert_text("P(A|B)", P(Distribution((A,), (B,))))
        self.assert_text("P(A|B)", P(A | [B]))
        self.assert_text("P(A|B,C)", P(Distribution((A,), (B,)) | C))
        self.assert_text("P(A|B,C)", P(A | [B, C]))
        self.assert_text("P(A|B,C)", P(A | B | C))
        self.assert_text("P(A|B,C)", P(A | B & C))

        # Test simple joint distributions
        self.assert_text("P(A,B)", P((A, B)))
        self.assert_text("P(A,B)", P(A, B))
        self.assert_text("P(A,B)", P(A & B))
        self.assert_text("P(A,B,C)", P(A & B & C))

        # Test mixed with single conditional
        self.assert_text("P(A,B|C)", P(Distribution((A, B), (C,))))
        self.assert_text("P(A,B|C)", P(Distribution((A, B), (C,))))
        self.assert_text("P(A,B|C)", P(Distribution((A, B)) | C))
        self.assert_text("P(A,B|C)", P(A & B | C))

        # Test mixed with multiple conditionals
        self.assert_text("P(A,B|C,D)", P(Distribution((A, B), (C, D))))
        self.assert_text("P(A,B|C,D)", P(Distribution((A, B)) | C | D))
        self.assert_text("P(A,B|C,D)", P(Distribution((A, B), (C,)) | D))
        self.assert_text("P(A,B|C,D)", P(A & B | C | D))
        self.assert_text("P(A,B|C,D)", P(A & B | (C, D)))
        self.assert_text("P(A,B|C,D)", P(A & B | Distribution((C, D))))
        self.assert_text("P(A,B|C,D)", P(A & B | C & D))

    def test_conditioning_errors(self):
        """Test erroring on conditionals."""
        for expression in [
            Distribution((B,), (C,)),
            Distribution((B, C), (D,)),
            Distribution((B, C), (D, W)),
        ]:
            with self.assertRaises(TypeError):
                _ = A | expression
            with self.assertRaises(TypeError):
                _ = X & Y | expression

    def test_sum(self):
        """Test the Sum DSL object."""
        # Sum with no variables
        self.assert_text(
            "[ sum_{} P(A|B) P(C|D) ]",
            Sum(P(A | B) * P(C | D)),
        )
        # Sum with one variable
        self.assert_text(
            "[ sum_{S} P(A|B) P(C|D) ]",
            Sum(P(A | B) * P(C | D), (S,)),
        )
        # Sum with two variables
        self.assert_text(
            "[ sum_{S,T} P(A|B) P(C|D) ]",
            Sum(P(A | B) * P(C | D), (S, T)),
        )

        # CRAZY sum syntax! pycharm doesn't like this usage of __class_getitem__ though so idk if we'll keep this
        self.assert_text(
            "[ sum_{S} P(A|B) P(C|D) ]",
            Sum[S](P(A | B) * P(C | D)),
        )
        self.assert_text(
            "[ sum_{S,T} P(A|B) P(C|D) ]",
            Sum[S, T](P(A | B) * P(C | D)),
        )

        # Sum with sum inside
        self.assert_text(
            "[ sum_{S,T} P(A|B) [ sum_{R} P(C|D) ] ]",
            Sum(P(A | B) * Sum(P(C | D), (R,)), (S, T)),
        )

    def test_q(self):
        """Test the Q DSL object."""
        self.assert_text("Q[A](X)", Q[A](X))
        self.assert_text("Q[A,B](X)", Q[A, B](X))
        self.assert_text("Q[A](X,Y)", Q[A](X, Y))
        self.assert_text("Q[A,B](X,Y)", Q[A, B](X, Y))

    def test_jeremy(self):
        """Test assorted complicated objects from Jeremy."""
        self.assert_text(
            "[ sum_{W} P(Y_{Z*,W},X) P(D) P(Z_{D}) P(W_{X*}) ]",
            Sum(P((Y @ ~Z @ W) & X) * P(D) * P(Z @ D) * P(W @ ~X), (W,)),
        )

        self.assert_text(
            "[ sum_{W} P(Y_{Z*,W},X) P(W_{X*}) ]",
            Sum(P(Y @ ~Z @ W & X) * P(W @ ~X), (W,)),
        )

        self.assert_text(
            "frac_{[ sum_{W} P(Y_{Z,W},X) P(W_{X*}) ]}{[ sum_{Y} [ sum_{W} P(Y_{Z,W},X) P(W_{X*}) ] ]}",
            Fraction(
                Sum(P(Y @ Z @ W & X) * P(W @ ~X), (W,)),
                Sum(Sum(P(Y @ Z @ W & X) * P(W @ ~X), (W,)), (Y,)),
            ),
        )

        self.assert_text(
            "[ sum_{D} P(Y_{Z*,W},X) P(D) P(Z_{D}) P(W_{X*}) ]",
            Sum(P(Y @ ~Z @ W & X) * P(D) * P(Z @ D) * P(W @ ~X), (D,)),
        )

        self.assert_text(
            "[ sum_{W,D,Z,V} [ sum_{} P(W|X) ] [ sum_{} [ sum_{X,W,Z,Y,V} P(X,W,D,Z,Y,V) ] ]"
            " [ sum_{} P(Z|D,V) ] [ sum_{} [ sum_{X} P(Y|X,D,V,Z,W) P(X) ] ]"
            " [ sum_{} [ sum_{X,W,D,Z,Y} P(X,W,D,Z,Y,V) ] ] ]",
            Sum[W, D, Z, V](
                Sum(P(W | X))
                * Sum(Sum[X, W, Z, Y, V](P(X, W, D, Z, Y, V)))
                * Sum(P(Z | [D, V]))
                * Sum(Sum[X](P(Y | [X, D, V, Z, W]) * P(X)))
                * Sum(Sum[X, W, D, Z, Y](P(X, W, D, Z, Y, V))),
            ),
        )

        """
        [[sum_{D,Z,V} [sum_{} [sum_{X,W,Z,Y,V} P(X,W,D,Z,Y,V)]][sum_{}P(Z|D,V)][sum_{} [sum_{X} P(Y|X,D,V,Z,W)P(X|)]]
        [sum_{} [sum_{X,W,D,Z,Y} P(X,W,D,Z,Y,V)]]]]/[ sum_{Y}[sum_{D,Z,V} [sum_{} [sum_{X,W,Z,Y,V} P(X,W,D,Z,Y,V)]]
        [sum_{}P(Z|D,V)][sum_{} [sum_{X} P(Y|X,D,V,Z,W)P(X|)]][sum_{} [sum_{X,W,D,Z,Y} P(X,W,D,Z,Y,V)]]]]
        """

    def test_api(self):
        """Test the high-level API for getting variables."""
        for expression, variables in [
            (A @ B, {A @ B, -B}),
            (A @ ~B, {A @ ~B, ~B}),
            (P(A @ B), {A @ B, -B}),
            (P(A @ ~B), {A @ ~B, ~B}),
            (P(A @ ~B) * P(C), {A @ ~B, ~B, C}),
            (Sum[S](P(A @ ~B) * P(C)), {A @ ~B, ~B, C, S}),
            (Sum[S, T](P(A @ ~B) * P(C)), {A @ ~B, ~B, C, S, T}),
            (One() / P(A @ B), {A @ B, -B}),
            (P(B) / P(A @ ~B), {A @ ~B, B, ~B}),
            (P(Y | X) * P(X) / P(Y), {X, Y}),
            (Q[A, B](C, D), {A, B, C, D}),
        ]:
            with self.subTest(expression=str(expression)):
                self.assertEqual(variables, expression.get_variables())



# -*- coding: utf-8 -*-

"""Data structures."""

from __future__ import annotations

from typing import Iterable, NamedTuple, Optional, Tuple

from .dsl import Expression, Variable

__all__ = [
    "VermaConstraint",
    "DSeparationJudgement",
]


class VermaConstraint(NamedTuple):
    """Represent a Verma constraint."""

    lhs_cfactor: Expression
    lhs_expr: Expression
    rhs_cfactor: Expression
    rhs_expr: Expression
    variables: Tuple[Variable, ...]

    @classmethod
    def from_element(cls, element) -> VermaConstraint:
        """Extract content from each element in the vector returned by `verma.constraint`.

        :param element: An element in the in the vector returned by `verma.constraint`
        :returns: A Verma constraint tuple for the given element

        .. seealso:: Extracting from R objects https://rpy2.github.io/doc/v3.4.x/html/vector.html#extracting-items
        """
        from .parser import parse_causaleffect
        from .r_utils import _extract, _parse_vars

        print(element)
        return cls(
            rhs_cfactor=parse_causaleffect(_extract(element, "rhs.cfactor")),
            rhs_expr=parse_causaleffect(_extract(element, "rhs.expr")),
            lhs_cfactor=parse_causaleffect(_extract(element, "lhs.cfactor")),
            lhs_expr=parse_causaleffect(_extract(element, "lhs.expr")),
            variables=_parse_vars(element),
        )


class DSeparationJudgement(NamedTuple):
    """
    Record if a left/right pair are d-separated given the conditions.

    By default, acts like a boolean, but also caries evidence graph.
    """

    separated: Optional[bool]
    left: str
    right: str
    conditions: Tuple[str, ...]

    @classmethod
    def create(
        cls,
        left: str,
        right: str,
        conditions: Optional[Iterable[str]] = None,
        *,
        separated: Optional[bool] = True,
    ) -> DSeparationJudgement:
        """Create a d-separation judgement in canonical form."""
        left, right = sorted([left, right])
        if conditions is None:
            conditions = tuple()
        conditions = tuple(sorted(set(conditions)))
        return cls(separated, left, right, conditions)

    def __bool__(self):
        return self.separated

    def __repr__(self):
        return f"{repr(self.separated)} -- '{self.left}' d-sep '{self.right}' conditioned on {self.conditions}"

    def __eq__(self, other):
        return self.separated == other

    @property
    def is_canonical(self) -> bool:
        """Return if the conditional independency is in canonical form."""
        return (
            self.left < self.right
            and isinstance(self.conditions, tuple)
            and tuple(sorted(self.conditions)) == (self.conditions)
        )



# -*- coding: utf-8 -*-

"""Tests for the CausalEffect wrapper."""

import unittest

import pyparsing

from y0.dsl import P, Q, Sum, Variable
from y0.examples import examples, verma_1

try:
    from y0.causaleffect import r_get_verma_constraints
    from y0.struct import VermaConstraint
    from y0.r_utils import CAUSALEFFECT, IGRAPH
except ImportError:  # rpy2 is not installed
    missing_rpy2 = True
else:
    missing_rpy2 = False

u_1 = Variable("u_1")
V1 = Variable("V1")
V2 = Variable("V2")
V3 = Variable("V3")
V4 = Variable("V4")


@unittest.skipIf(missing_rpy2, "rpy2 is not installed")
class TestCausalEffect(unittest.TestCase):
    """Tests for the causaleffect wrapper."""

    @classmethod
    def setUpClass(cls) -> None:
        """Make imports for the class."""
        from rpy2.robjects.packages import importr
        from rpy2.robjects.packages import PackageNotInstalledError

        try:
            importr(CAUSALEFFECT)
            importr(IGRAPH)
        except PackageNotInstalledError:
            raise unittest.SkipTest("R packages not properly installed.")

    def test_verma_constraint(self):
        """Test getting the single Verma constraint from the Figure 1A graph."""
        for example in examples:
            with self.subTest(name=example.name):
                try:
                    actual = r_get_verma_constraints(example.graph)
                except pyparsing.ParseException:
                    continue
                expected = example.verma_constraints
                self.assertEqual(set(expected or ()), set(actual))

        actual = r_get_verma_constraints(verma_1)
        self.assertEqual(1, len(actual))
        verma_constraint = actual[0]
        self.assertIsInstance(verma_constraint, VermaConstraint)
        self.assertEqual(Q[V4](V3, V4), verma_constraint.rhs_cfactor)

        expected_rhs_expr = Sum[u_1, V3](P(V4 | u_1 | V3) * P(V3) * P(u_1))
        self.assertEqual(
            expected_rhs_expr,
            verma_constraint.rhs_expr,
            msg=f"Expected: {expected_rhs_expr}\nActual:  {verma_constraint.rhs_expr}",
        )
        self.assertEqual(
            Sum[V2](Q[V2, V4](V1, V2, V3, V4)), verma_constraint.lhs_cfactor
        )
        self.assertEqual(
            Sum[V2](P(V4 | (V1, V2, V3)) * P(V2 | V1)), verma_constraint.lhs_expr
        )
        self.assertEqual((V1,), verma_constraint.variables)



# -*- coding: utf-8 -*-

"""Utilities for the parser for Craig-like probability expressions."""

from pyparsing import (
    Group,
    Optional,
    ParseResults,
    Suppress,
    Word,
    alphas,
    delimitedList,
)

from ...dsl import (
    CounterfactualVariable,
    Distribution,
    Intervention,
    Probability,
    QFactor,
    Variable,
)

__all__ = [
    "probability_pe",
    "variables_pe",
    "qfactor_pe",
]


def _parse_variable(_s, _l, tokens: ParseResults):
    return Variable(tokens[0])


letter = Word(alphas)


def _set_star(_s, _l, tokens: ParseResults):
    return 1 == len(tokens.asList())


def _make_intervention(_s, _l, tokens: ParseResults):
    return Intervention(name=tokens["name"], star=bool(tokens["star"]))


def _unpack(_s, _l, tokens: ParseResults):
    return tokens[0]


def _make_variable(_s, _l, tokens: ParseResults) -> Variable:
    if "interventions" not in tokens:
        return Variable(name=tokens["name"])
    return CounterfactualVariable(
        name=tokens["name"],
        interventions=tuple(tokens["interventions"].asList()),
    )


def _make_probability(_s, _l, tokens: ParseResults) -> Probability:
    children, parents = tokens["children"].asList(), tokens["parents"].asList()
    if not children:
        raise ValueError
    return Probability(Distribution(children=tuple(children), parents=tuple(parents)))


def _make_q(_s, _l, tokens: ParseResults) -> QFactor:
    codomain, domain = tokens["codomain"].asList(), tokens["domain"].asList()
    return QFactor(codomain=tuple(codomain), domain=tuple(domain))


# The suffix "pe" refers to :class:`pyparsing.ParserElement`, which is the
#  class in pyparsing that everything inherits from
star_pe = Optional("*")("star").setParseAction(_set_star)
intervention_pe = letter("name") + star_pe
intervention_pe.setParseAction(_make_intervention)
interventions_pe = Optional(
    Suppress("_{")
    + delimitedList(Group(intervention_pe).setParseAction(_unpack)).setResultsName(
        "interventions"
    )
    + Suppress("}"),
)
variable_pe = letter("name") + interventions_pe
variable_pe.setParseAction(_make_variable)
variable_pe.setName("variable")

variables_pe = delimitedList(Group(variable_pe).setParseAction(_unpack))
_children_pe = Group(variables_pe).setResultsName("children")
_parents_pe = Group(Optional(Suppress("|") + variables_pe)).setResultsName("parents")
probability_pe = Suppress("P(") + _children_pe + _parents_pe + Suppress(")")
probability_pe.setParseAction(_make_probability)
probability_pe.setName("probability")

qfactor_pe = (
    Suppress("Q[")
    + Group(variables_pe).setResultsName("codomain")
    + Suppress("](")
    + Group(variables_pe).setResultsName("domain")
    + Suppress(")")
)
qfactor_pe.setParseAction(_make_q)
qfactor_pe.setName("qfactor")
